# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Handle Ansible collections.
"""

from __future__ import annotations

import os
import shutil
import tempfile
import typing as t
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path

from antsibull_fileutils.yaml import load_yaml_file

from ..ansible import AnsibleCoreVersion
from ..paths import copy_collection as _paths_copy_collection
from ..paths import remove_path as _remove
from .data import CollectionData, CollectionSource, SetupResult
from .extract import extract_tarball
from .search import (
    CollectionList,
    _update_collection_list,
    get_collection_list,
)

# Function that runs a command (and fails on non-zero return code)
# and returns a tuple (stdout, stderr)
Runner = t.Callable[[list[str]], tuple[bytes, bytes]]


class _CollectionSources:
    sources: dict[str, CollectionSource]

    def __init__(self):
        self.sources = {}

    def set_source(self, name: str, source: CollectionSource) -> None:
        """
        Set source for collection.
        """
        self.sources[name] = source

    @t.overload
    def get_source(
        self, name: str, *, create_default: t.Literal[True] = True
    ) -> CollectionSource: ...

    @t.overload
    def get_source(
        self, name: str, *, create_default: t.Literal[False]
    ) -> CollectionSource | None: ...

    def get_source(
        self, name: str, *, create_default: bool = True
    ) -> CollectionSource | None:
        """
        Get source for collection.
        """
        source = self.sources.get(name)
        if source is None and create_default:
            source = CollectionSource(name, name)
        return source


class _CollectionDownloadCache:
    @staticmethod
    def _parse_galaxy_filename(file: Path) -> tuple[str, str, str]:
        """
        Split filename into (namespace, name, version) tuple.
        """
        if not file.name.endswith(_TARBALL_EXTENSION):
            raise ValueError(
                f"Filename {file.name!r} does not end with {_TARBALL_EXTENSION}"
            )
        parts = file.name[: -len(_TARBALL_EXTENSION)].split("-", 2)
        if len(parts) != 3:
            raise ValueError(
                f"Filename {file.name!r} does not belong to a Galaxy tarball"
            )
        return parts[0], parts[1], parts[2]

    @staticmethod
    def _parse_cache_filename(file: Path) -> tuple[str, str, str, str] | None:
        """
        Split cache filename into (namespace, name, source_id, version) tuple.
        """
        if not file.name.endswith(_TARBALL_EXTENSION):
            return None
        parts = file.name[: -len(_TARBALL_EXTENSION)].split("-", 3)
        if len(parts) != 4:
            return None
        return parts[0], parts[1], parts[2], parts[3]

    @staticmethod
    def _encode_cache_filename(
        namespace: str, name: str, version: str, source: CollectionSource
    ) -> str:
        return f"{namespace}-{name}-{source.identifier()}-{version}{_TARBALL_EXTENSION}"

    def download_collections(
        self, *, destination: Path, sources: list[CollectionSource], runner: Runner
    ) -> None:
        """
        Given a set of collection sources, downloads these and stores them in destination.
        """
        destination.mkdir(exist_ok=True)
        names = ", ".join(sorted(source.name for source in sources))
        print(f"Downloading {names} to {destination}...")
        sources_by_name = {}
        for source in sources:
            sources_by_name[source.name] = source
            if source.name != source.source:
                print(f"  Installing {source.name} via {source.source}...")
        with tempfile.TemporaryDirectory(
            prefix="antsibull-nox-galaxy-download"
        ) as dest:
            tempdir = Path(dest)
            command = [
                "ansible-galaxy",
                "collection",
                "download",
                "--no-deps",
                "--download-path",
                str(tempdir),
                "--",
                *(source.source for source in sources),
            ]
            runner(command)
            for file in tempdir.iterdir():
                if file.name.endswith(_TARBALL_EXTENSION) and file.is_file():
                    namespace, name, version = self._parse_galaxy_filename(file)
                    source_opt = sources_by_name.get(f"{namespace}.{name}")
                    if source_opt is None:
                        print(
                            f"Found unknown collection artifact {file.name!r}, ignoring..."
                        )
                        continue
                    destfile = destination / self._encode_cache_filename(
                        namespace, name, version, source_opt
                    )
                    _remove(destfile)
                    shutil.move(file, destfile)

    def list_downloaded_dir(
        self, *, path: Path
    ) -> dict[tuple[str, str], tuple[Path, str]]:
        """
        List contents of download cache.

        Returns a dictionary mapping (collection_name, source_id) tuples to
        (tarball_path, version) tuples.
        """
        if not path.is_dir():
            return {}
        result: dict[tuple[str, str], tuple[Path, str]] = {}
        for file in path.iterdir():
            if not file.is_file():
                continue
            parsed = self._parse_cache_filename(file)
            if parsed is None:
                continue
            namespace, name, source_id, version = parsed
            collection_name = f"{namespace}.{name}"
            key = collection_name, source_id
            if key in result:
                # Determine older entry
                old_file = result[key][0]
                old_stat = old_file.stat()
                new_stat = file.stat()
                if new_stat.st_mtime > old_stat.st_mtime:
                    older_file = old_file
                    result[key] = file, version
                else:
                    older_file = file
                # Clean up older entry
                _remove(older_file)
            else:
                result[key] = file, version
        return result


_COLLECTION_SOURCES = _CollectionSources()
_COLLECTION_SOURCES_PER_CORE_VERSION: dict[AnsibleCoreVersion, _CollectionSources] = {}
_COLLECTION_DOWNLOAD_CACHE = _CollectionDownloadCache()
_TARBALL_EXTENSION = ".tar.gz"
_INSTALLATION_CONFIG_ENV_VAR = "ANTSIBULL_NOX_INSTALL_COLLECTIONS"


def setup_collection_sources(
    collection_sources: dict[str, CollectionSource],
    *,
    ansible_core_version: AnsibleCoreVersion | None = None,
) -> None:
    """
    Setup collection sources.
    """
    sources = _COLLECTION_SOURCES
    if ansible_core_version is not None:
        sources = _COLLECTION_SOURCES_PER_CORE_VERSION.get(
            ansible_core_version, _COLLECTION_SOURCES
        )
        if sources is _COLLECTION_SOURCES:
            sources = _CollectionSources()
            _COLLECTION_SOURCES_PER_CORE_VERSION[ansible_core_version] = sources
    for name, source in collection_sources.items():
        sources.set_source(name, source)


def _install_from_download_cache(
    *, full_name: str, tarball: Path, destination: Path
) -> Path:
    destination_dir = destination / full_name
    _remove(destination_dir)
    print(f"Installing {full_name} from {tarball} to {destination_dir}...")
    extract_tarball(tarball=tarball, destination=destination_dir)
    return destination_dir


def _get_source(
    name: str, *, ansible_core_version: AnsibleCoreVersion
) -> CollectionSource:
    sources_per_version = _COLLECTION_SOURCES_PER_CORE_VERSION.get(ansible_core_version)
    if sources_per_version:
        result = sources_per_version.get_source(name, create_default=False)
        if result is not None:
            return result
    return _COLLECTION_SOURCES.get_source(name)


def _install_missing(
    collections: list[str],
    *,
    ansible_core_version: AnsibleCoreVersion,
    runner: Runner,
) -> list[CollectionData]:
    config = os.environ.get(_INSTALLATION_CONFIG_ENV_VAR)
    if config == "never":
        names = ", ".join(sorted(collections))
        plural_s = "" if len(collections) == 1 else "s"
        print(
            f"{_INSTALLATION_CONFIG_ENV_VAR} is set to 'never',"
            f" thus cannot install missing exception{plural_s} {names}..."
        )
        return []
    sources = [
        _get_source(name, ansible_core_version=ansible_core_version)
        for name in collections
    ]
    result: list[CollectionData] = []
    with _update_collection_list(ansible_core_version=ansible_core_version) as updater:
        global_cache = updater.get_global_cache()
        install: list[CollectionSource] = []
        download: list[CollectionSource] = []
        download_cache = _COLLECTION_DOWNLOAD_CACHE.list_downloaded_dir(
            path=global_cache.download_cache
        )
        for source in sources:
            if cd := updater.find(source.name):
                result.append(cd)
            else:
                install.append(source)
                if not download_cache.get((source.name, source.identifier())):
                    download.append(source)
        if download:
            _COLLECTION_DOWNLOAD_CACHE.download_collections(
                destination=global_cache.download_cache, sources=download, runner=runner
            )
            download_cache = _COLLECTION_DOWNLOAD_CACHE.list_downloaded_dir(
                path=global_cache.download_cache
            )
        if install:
            for source in install:
                key = source.name, source.identifier()
                if key not in download_cache:
                    raise ValueError(
                        f"Error: cannot find {source.name} (source ID {source.identifier()})"
                        f" in download cache {global_cache.download_cache}"
                        " after successful download!"
                    )
                c_dir = _install_from_download_cache(
                    full_name=source.name,
                    tarball=download_cache[key][0],
                    destination=global_cache.get_extracted_path(
                        ansible_core_version=ansible_core_version
                    ),
                )
                c_namespace, c_name = source.name.split(".", 1)
                result.append(
                    updater.add_collection(
                        directory=c_dir, namespace=c_namespace, name=c_name
                    )
                )
    return result


@dataclass(frozen=True, order=True)
class _Source:
    """
    Represents the source of a missing dependency.
    """

    name: str | None = None
    path: Path | None = None
    what: str | None = None

    @classmethod
    def dependency_of(cls, name: str) -> _Source:
        """
        Dependency of collection.
        """
        return cls(name=name)

    @classmethod
    def from_file(cls, path: Path) -> _Source:
        """
        Dependency from collection requirements file.
        """
        return cls(path=path)

    @classmethod
    def from_other(cls, what: str) -> _Source:
        """
        Another source.
        """
        return cls(what=what)

    def nice_str(self) -> str:
        """
        Convert to a nice (human readable) string.
        """
        if self.name:
            return f"dependency of {self.name}"
        if self.path:
            return f"required in {self.path}"
        if self.what:
            return f"required through {self.what}"
        return "(unknown)"


class _MissingDependency:
    """
    Models a missing dependency with a list of sources where it is required from.
    """

    name: str
    sources: set[_Source]

    def __init__(self, name: str, source: _Source) -> None:
        """
        Create missing dependency with source.
        """
        self.name = name
        self.sources = {source}

    def add_source(self, source: _Source) -> None:
        """
        Add source.
        """
        self.sources.add(source)


class _MissingDependencies:
    """
    Models all missing dependencies.
    """

    missing: dict[str, _MissingDependency]

    def __init__(self) -> None:
        self.missing = {}

    def is_empty(self) -> bool:
        """
        Query whether no collections are missing.
        """
        return not self.missing

    def get_missing_names(self) -> list[str]:
        """
        Get a sorted list of missing collections.
        """
        return sorted(self.missing)

    def add(self, name: str, *, source: _Source) -> None:
        """
        Add a missing dependency.
        """
        if name in self.missing:
            self.missing[name].add_source(source)
        else:
            self.missing[name] = _MissingDependency(name, source)

    def remove(self, name: str) -> None:
        """
        Remove a missing dependency (because it was installed).
        """
        self.missing.pop(name)

    def raise_error(self) -> None:
        """
        Raise a human-readable error about missing collections.
        If no collections are missing, simply return.
        """
        if not self.missing:
            return
        collections: list[str] = []
        for collection in sorted(self.missing):
            sources = sorted(self.missing[collection].sources)
            sources_text = ", ".join(source.nice_str() for source in sources)
            collections.append(f"{collection}  (required from {sources_text})")
        plural_s = "" if len(collections) == 1 else ""
        enumeration = "- " + "\n- ".join(collections)
        raise ValueError(
            f"The following collection{plural_s} are missing:\n{enumeration}"
        )


def _add_all_dependencies(
    collections: dict[str, CollectionData],
    missing_dependencies: _MissingDependencies,
    all_collections: CollectionList,
) -> None:
    to_process = list(collections.values())
    while to_process:
        collection = to_process.pop(0)
        for dependency_name in collection.dependencies:
            if dependency_name not in collections:
                dependency_data = all_collections.find(dependency_name)
                if dependency_data is None:
                    missing_dependencies.add(
                        dependency_name,
                        source=_Source.dependency_of(collection.full_name),
                    )
                    continue
                collections[dependency_name] = dependency_data
                to_process.append(dependency_data)


def _install_collection(collection: CollectionData, path: Path) -> None:
    # Compute relative path
    sym_path = collection.path.absolute().relative_to(path.parents[0], walk_up=True)
    # Ensure that path is symlink with this relative path
    if path.is_symlink():
        if path.readlink() == sym_path:
            return
        path.unlink()
    else:
        _remove(path)
    path.symlink_to(sym_path)


def _install_current_collection(collection: CollectionData, path: Path) -> None:
    if path.exists() and (path.is_symlink() or not path.is_dir()):
        path.unlink()
    path.mkdir(exist_ok=True)
    present = {p.name for p in path.iterdir()}
    for source_entry in collection.path.absolute().iterdir():
        if source_entry.name == ".nox":
            continue
        dest_entry = path / source_entry.name
        # Compute relative path
        sym_path = source_entry.relative_to(path, walk_up=True)
        # Ensure that dest_entry is symlink with this relative path
        if source_entry.name in present:
            present.remove(source_entry.name)
            if dest_entry.is_symlink() and dest_entry.readlink() == sym_path:
                continue
            _remove(dest_entry)
        dest_entry.symlink_to(sym_path)
    for name in present:
        dest_entry = path / name
        _remove(dest_entry)


def _install_collections(
    collections: Iterable[CollectionData], root: Path, *, with_current: bool
) -> None:
    for collection in collections:
        namespace_dir = root / collection.namespace
        namespace_dir.mkdir(exist_ok=True)
        path = namespace_dir / collection.name
        if not collection.current:
            _install_collection(collection, path)
        elif with_current:
            _install_current_collection(collection, path)


def _extract_collections_from_extra_deps_file(path: str | os.PathLike) -> list[str]:
    if not os.path.exists(path):
        return []
    try:
        data = load_yaml_file(path)
        result = []
        if data.get("collections"):
            for index, collection in enumerate(data["collections"]):
                if isinstance(collection, str):
                    result.append(collection)
                    continue
                if not isinstance(collection, dict):
                    raise ValueError(
                        f"Collection entry #{index + 1} must be a string or dictionary"
                    )
                if not isinstance(collection.get("name"), str):
                    raise ValueError(
                        f"Collection entry #{index + 1} does not have a 'name' field of type string"
                    )
                result.append(collection["name"])
        return result
    except Exception as exc:
        raise ValueError(
            f"Error while loading collection dependency file {path}: {exc}"
        ) from exc


def setup_collections(
    destination: str | os.PathLike,
    runner: Runner,
    *,
    ansible_core_version: AnsibleCoreVersion,
    extra_collections: list[str] | None = None,
    extra_deps_files: list[str | os.PathLike] | None = None,
    global_cache_dir: Path,
    with_current: bool = True,
) -> SetupResult:
    """
    Setup all collections in a tree structure inside the destination directory.
    """
    all_collections = get_collection_list(
        runner=runner,
        global_cache_dir=global_cache_dir,
        ansible_core_version=ansible_core_version,
    )
    destination_root = Path(destination) / "ansible_collections"
    destination_root.mkdir(exist_ok=True)
    current = all_collections.current
    collections_to_install = {current.full_name: current}
    missing = _MissingDependencies()
    if extra_collections:
        for collection in extra_collections:
            collection_data = all_collections.find(collection)
            if collection_data is None:
                missing.add(collection, source=_Source.from_other("noxfile"))
            else:
                collections_to_install[collection_data.full_name] = collection_data
    if extra_deps_files is not None:
        for extra_deps_file in extra_deps_files:
            path = Path(extra_deps_file)
            for collection in _extract_collections_from_extra_deps_file(path):
                collection_data = all_collections.find(collection)
                if collection_data is None:
                    missing.add(collection, source=_Source.from_file(path))
                else:
                    collections_to_install[collection_data.full_name] = collection_data
    while True:
        _add_all_dependencies(collections_to_install, missing, all_collections)
        if missing.is_empty():
            break
        for collection_data in _install_missing(
            missing.get_missing_names(),
            ansible_core_version=ansible_core_version,
            runner=runner,
        ):
            collections_to_install[collection_data.full_name] = collection_data
            missing.remove(collection_data.full_name)
        missing.raise_error()
    _install_collections(
        collections_to_install.values(), destination_root, with_current=with_current
    )
    return SetupResult(
        root=destination_root,
        current_collection=current,
        current_path=(
            (destination_root / current.namespace / current.name)
            if with_current
            else None
        ),
    )


def _copy_collection(collection: CollectionData, path: Path) -> None:
    _paths_copy_collection(collection.path, path)


def _copy_collection_rsync_hard_links(
    collection: CollectionData, path: Path, runner: Runner
) -> None:
    _, __ = runner(
        [
            "rsync",
            "-av",
            "--delete",
            "--exclude",
            ".nox",
            "--link-dest",
            str(collection.path) + "/",
            "--",
            str(collection.path) + "/",
            str(path) + "/",
        ]
    )


def setup_current_tree(
    place: str | os.PathLike, current_collection: CollectionData
) -> SetupResult:
    """
    Setup a tree structure with the current collection in it.
    """

    path = Path(place)
    root = path / "ansible_collections"
    root.mkdir(exist_ok=True)
    namespace = root / current_collection.namespace
    namespace.mkdir(exist_ok=True)
    collection = namespace / current_collection.name
    _copy_collection(current_collection, collection)
    # _copy_collection_rsync_hard_links(current_collection, collection, runner)
    return SetupResult(
        root=root,
        current_collection=current_collection,
        current_path=collection,
    )


__all__ = [
    "get_collection_list",
    "setup_collections",
    "setup_current_tree",
    "setup_collection_sources",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Python version utilities.
"""

from __future__ import annotations

import functools
import shutil
import subprocess
from pathlib import Path

from .utils import Version

# The following contains Python version candidates
_PYTHON_VERSIONS_TO_TRY: tuple[Version, ...] = tuple(
    Version.parse(v)
    for v in [
        # Python 2:
        "2.6",
        "2.7",
        # Python 3:
        "3.5",
        "3.6",
        "3.7",
        "3.8",
        "3.9",
        "3.10",
        "3.11",
        "3.12",
        "3.13",
        "3.14",
        # "3.15",
        # "3.16",
    ]
)


@functools.cache
def get_installed_python_versions() -> dict[Version, Path]:
    """
    Return a map of supported Python versions for which interpreters exist and are on the path,
    mapped to an executable.
    """
    result = {}

    # Look for pythonX.Y binaries
    for candidate in _PYTHON_VERSIONS_TO_TRY:
        if exe := shutil.which(f"python{candidate}"):
            result[candidate] = Path(exe)

    # Look for python, python2, python3 binaries and determine their version
    for executable in ("python", "python2", "python3"):
        exe = shutil.which(executable)
        if exe:
            script = "import platform; print('.'.join(platform.python_version().split('.')[:2]))"
            exe_result = subprocess.run(
                [exe, "-c", script],
                check=False,
                text=True,
                capture_output=True,
            )
            if exe_result.returncode == 0 and exe_result.stdout:
                try:
                    version = Version.parse(exe_result.stdout.strip())
                except (AttributeError, ValueError):
                    continue
                else:
                    result.setdefault(version, Path(exe))

    return result


__all__ = [
    "get_installed_python_versions",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

# pylint: disable=missing-function-docstring

"""
Tests for the collection module
"""

from __future__ import annotations

import typing as t
from pathlib import Path

import pytest
from antsibull_fileutils.yaml import load_yaml_file, store_yaml_file

from antsibull_nox.collection import (
    CollectionData,
    Runner,
    force_collection_version,
    load_collection_data_from_disk,
)
from antsibull_nox.collection.install import (
    _add_all_dependencies,
    _extract_collections_from_extra_deps_file,
    _MissingDependencies,
)
from antsibull_nox.collection.search import (
    _COLLECTION_LIST,
    CollectionList,
    _fs_list_local_collections,
    _galaxy_list_collections,
    get_collection_list,
)

from .utils import chdir


def create_once_runner(args: list[str], stdout: bytes, stderr: bytes = b"") -> Runner:
    call_counter = [0]

    def runner(call_args: list[str]) -> tuple[bytes, bytes]:
        assert call_counter[0] == 0
        assert call_args == args
        call_counter[0] += 1
        return stdout, stderr

    return runner


def test__add_all_dependencies() -> None:
    path = Path.cwd()
    all_collections = CollectionList.create(
        {
            c.full_name: c
            for c in [
                CollectionData.create(path=path, full_name="foo.bar", current=True),
                CollectionData.create(
                    path=path, full_name="foo.bam", dependencies={"foo.bar": "*"}
                ),
                CollectionData.create(
                    path=path,
                    full_name="foo.baz",
                    dependencies={"foo.bar": "*", "foo.bam": ">= 1.0.0"},
                ),
                CollectionData.create(
                    path=path, full_name="foo.foo", dependencies={"foo.bam": "*"}
                ),
                CollectionData.create(
                    path=path,
                    full_name="foo.error",
                    dependencies={"foo.does_not_exist": "*"},
                ),
            ]
        }
    )

    # No deps
    deps: dict[str, CollectionData] = {}
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert deps == {}  # pylint: disable=use-implicit-booleaness-not-comparison
    assert missing.is_empty()

    # Collection without deps
    deps = {
        name: all_collections.find(name)
        for name in [
            "foo.bar",
        ]
    }
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert deps.keys() == {"foo.bar"}
    assert missing.is_empty()

    # Collection with single dep
    deps = {
        name: all_collections.find(name)
        for name in [
            "foo.bam",
        ]
    }
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert deps.keys() == {"foo.bar", "foo.bam"}
    assert missing.is_empty()

    # Collection with two deps
    deps = {
        name: all_collections.find(name)
        for name in [
            "foo.baz",
        ]
    }
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert deps.keys() == {"foo.bar", "foo.bam", "foo.baz"}
    assert missing.is_empty()

    # Collection with dependency chain where leaf is already there
    deps = {
        name: all_collections.find(name)
        for name in [
            "foo.bar",
            "foo.foo",
        ]
    }
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert deps.keys() == {"foo.bar", "foo.bam", "foo.foo"}
    assert missing.is_empty()

    # Missing collection
    deps = {
        name: all_collections.find(name)
        for name in [
            "foo.error",
        ]
    }
    missing = _MissingDependencies()
    _add_all_dependencies(deps, missing, all_collections)
    assert not missing.is_empty()
    assert missing.get_missing_names() == ["foo.does_not_exist"]


def test__extract_collections_from_extra_deps_file_special(tmp_path: Path) -> None:
    # pylint: disable-next=use-implicit-booleaness-not-comparison
    assert _extract_collections_from_extra_deps_file(tmp_path / "does-not-exist") == []

    dir = tmp_path / "test1"
    dir.mkdir()
    with pytest.raises(
        ValueError,
        match="Error while loading collection dependency file.*Is a directory:",
    ):
        _extract_collections_from_extra_deps_file(dir)


EXTRACT_COLLECTIONS_FROM_EXTRA_DEPS_FILE_DATA: list[tuple[str, list[str]]] = [
    (
        r"""
collections: []
""",
        [],
    ),
    (
        r"""
collections:
    - foo
    - name: bar
""",
        ["foo", "bar"],
    ),
    (
        # Not exactly legal, but works:
        r"""
collections:
    foo: bar
    baz:
""",
        ["foo", "baz"],
    ),
]


@pytest.mark.parametrize(
    "content, expected_result",
    EXTRACT_COLLECTIONS_FROM_EXTRA_DEPS_FILE_DATA,
)
def test__extract_collections_from_extra_deps_file(
    content: str, expected_result: list[str], tmp_path: Path
) -> None:
    file = tmp_path / "test1"
    file.write_text(content)
    assert _extract_collections_from_extra_deps_file(file) == expected_result


EXTRACT_COLLECTIONS_FROM_EXTRA_DEPS_FILE_FAIL_DATA: list[tuple[str, str]] = [
    (
        r"""
collections:
    foo: bar
    23: baz
""",
        "Collection entry #2 must be a string or dictionary",
    ),
    (
        r"""
collections:
    - 42
""",
        "Collection entry #1 must be a string or dictionary",
    ),
    (
        r"""
collections:
    - name:
""",
        "Collection entry #1 does not have a 'name' field of type string",
    ),
    (
        r"""
collections:
    - bar: baz
""",
        "Collection entry #1 does not have a 'name' field of type string",
    ),
]


@pytest.mark.parametrize(
    "content, expected_match",
    EXTRACT_COLLECTIONS_FROM_EXTRA_DEPS_FILE_FAIL_DATA,
)
def test__extract_collections_from_extra_deps_file_fail(
    content: str, expected_match: str, tmp_path: Path
) -> None:
    file = tmp_path / "test1"
    file.write_text(content)
    with pytest.raises(ValueError, match=expected_match):
        _extract_collections_from_extra_deps_file(file)


def create_collection(
    path: Path,
    *,
    namespace: str,
    name: str,
    version: str | None = None,
    dependencies: dict[str, str] | None = None,
) -> None:
    data: dict[str, t.Any] = {
        "namespace": namespace,
        "name": name,
    }
    if version is not None:
        data["version"] = version
    if dependencies is not None:
        data["dependencies"] = dependencies
    path.mkdir(parents=True, exist_ok=True)
    store_yaml_file(path / "galaxy.yml", data)


def create_collection_w_dir(
    root: Path,
    *,
    namespace: str,
    name: str,
    version: str | None = None,
    dependencies: dict[str, str] | None = None,
) -> Path:
    path = root / namespace / name
    create_collection(
        path=path,
        namespace=namespace,
        name=name,
        version=version,
        dependencies=dependencies,
    )
    return path


def create_collection_w_shallow_dir(
    root: Path,
    *,
    directory_override: str | None = None,
    namespace: str,
    name: str,
    version: str | None = None,
    dependencies: dict[str, str] | None = None,
) -> Path:
    path = root / (directory_override or f"{namespace}.{name}")
    create_collection(
        path=path,
        namespace=namespace,
        name=name,
        version=version,
        dependencies=dependencies,
    )
    return path


def test__fs_list_local_collections(tmp_path: Path) -> None:
    # Case 1: regular ansible_collections tree
    root = tmp_path / "test-1" / "ansible_collections"
    root.mkdir(parents=True)
    foo_bar = create_collection_w_dir(root, namespace="foo", name="bar")
    foo_bam = create_collection_w_dir(
        root,
        namespace="foo",
        name="bam",
        dependencies={"foo.bar": ">= 1.0.0", "community.baz": "*"},
    )
    community_baz = create_collection_w_dir(
        root, namespace="community", name="baz", dependencies={"community.foo": "*"}
    )
    (root / "blah").write_text("nothing")
    (root / "blahsym").symlink_to("blah")
    (root / "empty").mkdir()
    (root / "blubb").mkdir()
    (root / "blubb" / "foo").write_text("nothing")
    (root / "blubb" / "foosym").symlink_to("foo")
    (root / "community" / "foo").write_text("nothing")
    (root / "community" / "bam").mkdir()
    with chdir(foo_bar):
        result = sorted(_fs_list_local_collections(), key=lambda c: c.full_name)
    assert result == [
        CollectionData.create(
            collections_root_path=root,
            path=community_baz,
            full_name="community.baz",
            dependencies={"community.foo": "*"},
        ),
        CollectionData.create(
            collections_root_path=root,
            path=foo_bam,
            full_name="foo.bam",
            dependencies={"foo.bar": ">= 1.0.0", "community.baz": "*"},
        ),
        CollectionData.create(
            collections_root_path=root, path=foo_bar, full_name="foo.bar", current=True
        ),
    ]

    # Case 2: repositories checked out as <namespace>.<name>
    root = tmp_path / "test-2"
    root.mkdir(parents=True)
    foo_bar = create_collection_w_shallow_dir(root, namespace="foo", name="bar")
    foo_bam = create_collection_w_shallow_dir(
        root,
        namespace="foo",
        name="bam",
        dependencies={"foo.bar": ">= 1.0.0", "community.baz": "*"},
    )
    create_collection_w_shallow_dir(
        root, namespace="bar", name="baz", directory_override="bar.bar"
    )
    community_baz = create_collection_w_shallow_dir(
        root, namespace="community", name="baz", dependencies={"community.foo": "*"}
    )
    (root / "blah").write_text("nothing")
    (root / "empty").mkdir()
    (root / "foo.baz").mkdir()
    (root / "1.2").mkdir()
    with chdir(foo_bar):
        result = sorted(_fs_list_local_collections(), key=lambda c: c.full_name)
    assert result == [
        CollectionData.create(
            path=community_baz,
            full_name="community.baz",
            dependencies={"community.foo": "*"},
        ),
        CollectionData.create(
            path=foo_bam,
            full_name="foo.bam",
            dependencies={"foo.bar": ">= 1.0.0", "community.baz": "*"},
        ),
        CollectionData.create(path=foo_bar, full_name="foo.bar", current=True),
    ]

    # Case 3: looks like ansible_collection tree on first glance, but it's not
    root = tmp_path / "test-3" / "ansible_collections" / "foo"
    root.mkdir(parents=True)
    foo_bar = create_collection_w_shallow_dir(root, namespace="foo", name="bar")
    foo_bam = create_collection_w_shallow_dir(
        root,
        namespace="foo",
        name="bam",
        dependencies={"foo.bar": ">= 1.0.0", "community.baz": "*"},
    )
    create_collection_w_shallow_dir(
        root, namespace="bar", name="baz", directory_override="bar.bar"
    )
    (root / "foo.baz").mkdir()
    with chdir(foo_bam):
        result = sorted(_fs_list_local_collections(), key=lambda c: c.full_name)
    assert result == [
        CollectionData.create(
            path=foo_bam,
            full_name="foo.bam",
            dependencies={"foo.bar": ">= 1.0.0", "community.baz": "*"},
            current=True,
        ),
        CollectionData.create(
            path=foo_bar,
            full_name="foo.bar",
        ),
    ]

    # Failure while loading current collection
    root = tmp_path / "test-4"
    root.mkdir(parents=True)
    cwd = root / "something"
    cwd.mkdir()
    with chdir(cwd):
        with pytest.raises(
            ValueError,
            match=(
                "^Cannot load current collection's info from.*:"
                " Cannot find galaxy.yml or MANIFEST.json in "
            ),
        ):
            list(_fs_list_local_collections())

        (cwd / "MANIFEST.json").write_text("foo")
        with pytest.raises(
            ValueError,
            match=(
                "^Cannot load current collection's info from.*"
                "/something: Cannot parse .*something/MANIFEST.json: "
            ),
        ):
            list(_fs_list_local_collections())

        (cwd / "MANIFEST.json").write_text("{}")
        with pytest.raises(
            ValueError,
            match=(
                "^Cannot load current collection's info from.*"
                "/something/MANIFEST.json does not contain collection_info$"
            ),
        ):
            list(_fs_list_local_collections())

        (cwd / "MANIFEST.json").write_text('{"collection_info": "meh"}')
        with pytest.raises(
            ValueError,
            match=(
                "^Cannot load current collection's info from.*"
                "/something/MANIFEST.json does not contain collection_info$"
            ),
        ):
            list(_fs_list_local_collections())

        (cwd / "MANIFEST.json").write_text('{"collection_info": {}}')
        with pytest.raises(
            ValueError,
            match=(
                "^Cannot load current collection's info from.*"
                "/something/MANIFEST.json does not contain a namespace$"
            ),
        ):
            list(_fs_list_local_collections())

        (cwd / "galaxy.yml").write_text("{")
        with pytest.raises(
            ValueError,
            match=(
                "^Cannot load current collection's info from.*"
                "/something: Cannot parse .*something/galaxy.yml: "
            ),
        ):
            list(_fs_list_local_collections())

        (cwd / "galaxy.yml").write_text("[]")
        with pytest.raises(
            ValueError,
            match=(
                "^Cannot load current collection's info from.*"
                "/something: .*something/galaxy.yml is not a dictionary"
            ),
        ):
            list(_fs_list_local_collections())

        (cwd / "galaxy.yml").write_text("namespace: whatever")
        with pytest.raises(
            ValueError,
            match=(
                "^Cannot load current collection's info from.*"
                "something/galaxy.yml does not contain a name$"
            ),
        ):
            list(_fs_list_local_collections())

        (cwd / "galaxy.yml").write_text("namespace: 42")
        with pytest.raises(
            ValueError,
            match=(
                "^Cannot load current collection's info from.*"
                "something/galaxy.yml does not contain a namespace$"
            ),
        ):
            list(_fs_list_local_collections())

        (cwd / "galaxy.yml").write_text(
            "namespace: foo\nname: bar\ndependencies: [foo]"
        )
        with pytest.raises(
            ValueError,
            match=(
                "^Cannot load current collection's info from.*"
                "something/galaxy.yml's dependencies is not a mapping$"
            ),
        ):
            list(_fs_list_local_collections())


GALAXY_LIST_COLLECTIONS_DATA: list[tuple[str, list[dict[str, t.Any]]]] = [
    (
        r"""
{}
""",
        [],
    ),
    (
        r"""
{
    "/": {}
}
""",
        [],
    ),
    (
        r"""
{
    "<ROOT1>": {
        "foo.bar": {}
    }
}
""",
        [
            {
                "root": 1,
                "full_name": "foo.bar",
            },
            {
                "root": 1,
                "full_name": "foo.bam",
                "hide": True,
            },
        ],
    ),
    (
        r"""
{
    "<ROOT1>": {
        "foo.bar": {},
        "foo.bam": {}
    },
    "<ROOT2>": {
        "foo.bam": {},
        "foo.baz": {}
    },
    "/does-not-exist": {
        "foo.bar": {}
    }
}
""",
        [
            {
                "root": 1,
                "full_name": "foo.bar",
            },
            {
                "root": 1,
                "full_name": "foo.bam",
            },
            {
                "root": 1,
                "full_name": "foo.baz",
                "hide": True,
            },
            {
                "root": 2,
                "full_name": "foo.bam",
                "version": "1.0.0",
            },
            {"root": 2, "full_name": "foo.baz", "dependencies": {"foo.bar": "*"}},
        ],
    ),
]


@pytest.mark.parametrize(
    "content, expected_result",
    GALAXY_LIST_COLLECTIONS_DATA,
)
def test__galaxy_list_collections(
    tmp_path: Path, content: str, expected_result: list[dict[str, t.Any]]
) -> None:
    root1 = tmp_path / "root-1" / "ansible_collections"
    root2 = tmp_path / "root-2" / "ansible_collections"
    root3 = tmp_path / "root-3" / "ansible_collections"

    def create(c: dict[str, t.Any]) -> tuple[CollectionData, bool]:
        root: Path = {
            1: root1,
            2: root2,
            3: root3,
        }[c.pop("root")]
        full_name: str = c.pop("full_name")
        hide: bool = c.pop("hide", False)
        namespace, name = full_name.split(".", 1)
        path = create_collection_w_dir(root, namespace=namespace, name=name, **c)
        return (
            CollectionData.create(
                collections_root_path=root,
                path=path,
                full_name=full_name,
                **c,
            ),
            hide,
        )

    expected_res_with_hide = [create(c) for c in expected_result]
    expected_res = sorted(
        [c for c, h in expected_res_with_hide if not h], key=lambda c: c.full_name
    )

    result = _galaxy_list_collections(
        create_once_runner(
            ["ansible-galaxy", "collection", "list", "--format", "json"],
            stdout=content.replace("<ROOT1>", str(root1))
            .replace("<ROOT2>", str(root2))
            .replace("<ROOT3>", str(root3))
            .encode("utf-8"),
        )
    )
    res = sorted(result, key=lambda c: c.full_name)
    assert res == expected_res


def test__galaxy_list_collections_fail() -> None:
    with pytest.raises(
        ValueError,
        match=(
            "^Error while loading collection list: "
            "Expecting property name enclosed in double quotes: "
        ),
    ):
        list(
            _galaxy_list_collections(
                create_once_runner(
                    ["ansible-galaxy", "collection", "list", "--format", "json"],
                    stdout=b"{",
                )
            )
        )


LOAD_COLLECTION_DATA_FROM_DISK_DATA: list[
    tuple[str, str, dict[str, t.Any], dict[str, t.Any]]
] = [
    (
        "galaxy.yml",
        r"""
namespace: foo
name: bar
version: 1.0.0
dependencies: {}
""",
        {},
        {
            "full_name": "foo.bar",
            "version": "1.0.0",
        },
    ),
    (
        "galaxy.yml",
        r"""
namespace: foo
name: bar
version: 1.0.0
dependencies:
    foo.bam: "*"
""",
        {
            "namespace": "foo",
            "name": "bar",
            "current": True,
            "root": "/",
        },
        {
            "full_name": "foo.bar",
            "version": "1.0.0",
            "dependencies": {"foo.bam": "*"},
            "current": True,
            "collections_root_path": "/",
        },
    ),
]


@pytest.mark.parametrize(
    "filename, content, paras, expected_result",
    LOAD_COLLECTION_DATA_FROM_DISK_DATA,
)
def test_load_collection_data_from_disk(
    filename: str,
    content: str,
    paras: dict[str, t.Any],
    expected_result: dict[str, t.Any],
    tmp_path: Path,
) -> None:
    (tmp_path / filename).write_text(content)
    res = load_collection_data_from_disk(tmp_path, **paras)
    expected_res = CollectionData.create(path=tmp_path, **expected_result)
    assert res == expected_res


LOAD_COLLECTION_DATA_FROM_DISK_FAIL_DATA: list[
    tuple[str, str, dict[str, t.Any], str]
] = [
    (
        "galaxy.yml",
        r"""
name: foo.bar
""",
        {},
        "/galaxy.yml does not contain a namespace$",
    ),
    (
        "galaxy.yml",
        r"""
namespace: foo
name: bar
version: 1.0.0
""",
        {
            "namespace": "foo",
            "name": "bam",
        },
        "/galaxy.yml contains name 'bar', but was hoping for 'bam'$",
    ),
    (
        "galaxy.yml",
        r"""
namespace: foo
name: bar
version: null
""",
        {
            "namespace": "fuu",
            "name": "bar",
        },
        "/galaxy.yml contains namespace 'foo', but was hoping for 'fuu'$",
    ),
    (
        "MANIFEST.json",
        "{}",
        {"accept_manifest": False},
        "^Cannot find galaxy.yml in ",
    ),
]


@pytest.mark.parametrize(
    "filename, content, paras, expected_match",
    LOAD_COLLECTION_DATA_FROM_DISK_FAIL_DATA,
)
def test_load_collection_data_from_disk_fail(
    filename: str,
    content: str,
    paras: dict[str, t.Any],
    expected_match: str,
    tmp_path: Path,
) -> None:
    (tmp_path / filename).write_text(content)
    with pytest.raises(ValueError, match=expected_match):
        load_collection_data_from_disk(tmp_path, **paras)


def test_get_collection_list(tmp_path) -> None:
    root1 = tmp_path / "root-1" / "ansible_collections"
    root2 = tmp_path / "root-2" / "ansible_collections"
    root3 = tmp_path / "root-3" / "ansible_collections"
    empty = tmp_path / "empty"

    create_collection_w_dir(root1, namespace="foo", name="bar", version="1.0.0")
    root1_foo_bam = create_collection_w_dir(
        root1, namespace="foo", name="bam", dependencies={"foo.bar": ">= 1.0.0"}
    )
    create_collection_w_dir(root2, namespace="foo", name="bam", version="0.1.0")
    root3_foo_bar = create_collection_w_dir(root3, namespace="foo", name="bar")

    content = r"""
{
    "<ROOT1>": {
        "foo.bar": {},
        "foo.bam": {}
    },
    "<ROOT2>": {
        "foo.bam": {},
        "foo.baz": {}
    },
    "<ROOT3>": {
        "foo.bar": {}
    }
}
"""

    runner = create_once_runner(
        ["ansible-galaxy", "collection", "list", "--format", "json"],
        stdout=content.replace("<ROOT1>", str(root1))
        .replace("<ROOT2>", str(root2))
        .replace("<ROOT3>", str(root3))
        .encode("utf-8"),
    )

    with chdir(root3 / "foo" / "bar"):
        _COLLECTION_LIST.clear()
        assert _COLLECTION_LIST.get_cached() is None
        result = get_collection_list(
            runner=runner, global_cache_dir=empty, ansible_core_version="devel"
        )
        cl = _COLLECTION_LIST.get_cached()
        assert cl is not None
        result_2 = get_collection_list(
            runner=runner, global_cache_dir=empty, ansible_core_version="devel"
        )
        assert _COLLECTION_LIST.get_cached() is cl

        with pytest.raises(
            ValueError, match="^Setup mismatch: global cache dir cannot be both "
        ):
            get_collection_list(
                runner=runner, global_cache_dir=tmp_path, ansible_core_version="devel"
            )

    assert result.collections == sorted(
        [
            CollectionData.create(
                collections_root_path=root1,
                path=root1_foo_bam,
                full_name="foo.bam",
                dependencies={"foo.bar": ">= 1.0.0"},
            ),
            CollectionData.create(
                collections_root_path=root3,
                path=root3_foo_bar,
                full_name="foo.bar",
                current=True,
            ),
        ],
        key=lambda c: c.full_name,
    )

    assert result == cl
    assert result == result_2


FORCE_COLLECTION_VERSION_DATA: list[tuple[str, str, bool, dict[str, t.Any]]] = [
    (
        r"""name: foo""",
        "1.2.3",
        True,
        {
            "name": "foo",
            "version": "1.2.3",
        },
    ),
    (
        r"""version: []""",
        "1.2.3",
        True,
        {
            "version": "1.2.3",
        },
    ),
    (
        r"""version: 1.2.2""",
        "1.2.3",
        True,
        {
            "version": "1.2.3",
        },
    ),
    (
        r"""version: 1.2.3""",
        "1.2.3",
        False,
        {
            "version": "1.2.3",
        },
    ),
    (
        r"""
name: boo
namespace: foo
version: null""",
        "1.2.3",
        True,
        {
            "name": "boo",
            "namespace": "foo",
            "version": "1.2.3",
        },
    ),
]


@pytest.mark.parametrize(
    "content, version, expected_result, expected_content",
    FORCE_COLLECTION_VERSION_DATA,
)
def test_force_collection_version(
    content: str,
    version: str,
    expected_result: bool,
    expected_content: dict[str, t.Any],
    tmp_path: Path,
) -> None:
    file = tmp_path / "galaxy.yml"
    file.write_text(content)
    result = force_collection_version(tmp_path, version=version)
    assert result == expected_result
    assert load_yaml_file(file) == expected_content


FORCE_COLLECTION_VERSION_FAIL_DATA: list[tuple[str, str, str]] = [
    (
        r"""{""",
        "1.2.3",
        "^Cannot parse .*/galaxy.yml: ",
    ),
]


@pytest.mark.parametrize(
    "content, version, expected_match",
    FORCE_COLLECTION_VERSION_FAIL_DATA,
)
def test_force_collection_version_fail(
    content: str, version: str, expected_match: str, tmp_path: Path
) -> None:
    file = tmp_path / "galaxy.yml"
    file.write_text(content)
    with pytest.raises(ValueError, match=expected_match):
        force_collection_version(tmp_path, version=version)



#!/usr/bin/env python

# Copyright (c) 2025, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Make sure all plugin and module documentation adheres to yamllint."""

from __future__ import annotations

import io
import os
import sys
import traceback
import typing as t

from antsibull_nox_data_util import setup  # type: ignore
from yamllint import linter
from yamllint.cli import find_project_config_filepath
from yamllint.config import YamlLintConfig
from yamllint.linter import PROBLEM_LEVELS

REPORT_LEVELS: set[PROBLEM_LEVELS] = {
    "warning",
    "error",
}


def lint(
    *,
    errors: list[dict[str, t.Any]],
    path: str,
    data: str,
    config: YamlLintConfig,
) -> None:
    try:
        problems = linter.run(
            io.StringIO(data),
            config,
            path,
        )
        for problem in problems:
            if problem.level not in REPORT_LEVELS:
                continue
            msg = f"{problem.level}: {problem.desc}"
            if problem.rule:
                msg += f"  ({problem.rule})"
            errors.append(
                {
                    "path": path,
                    "line": problem.line,
                    "col": problem.column,
                    "message": msg,
                }
            )
    except Exception as exc:
        error = str(exc).replace("\n", " / ")
        errors.append(
            {
                "path": path,
                "line": 1,
                "col": 1,
                "message": (
                    f"Internal error while linting YAML: exception {type(exc)}:"
                    f" {error}; traceback: {traceback.format_exc()!r}"
                ),
            }
        )


def process_yaml_file(
    errors: list[dict[str, t.Any]],
    path: str,
    config: YamlLintConfig,
) -> None:
    try:
        with open(path, "rt", encoding="utf-8") as stream:
            data = stream.read()
    except Exception as exc:
        errors.append(
            {
                "path": path,
                "line": 1,
                "col": 1,
                "message": (
                    f"Error while parsing Python code: exception {type(exc)}:"
                    f" {exc}; traceback: {traceback.format_exc()!r}"
                ),
            }
        )
        return

    lint(
        errors=errors,
        path=path,
        data=data,
        config=config,
    )


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()
    config: str | None = extra_data.get("config")

    if config is None:
        config = find_project_config_filepath()

    if config:
        yamllint_config = YamlLintConfig(file=config)
    else:
        yamllint_config = YamlLintConfig(content="extends: default")

    errors: list[dict[str, t.Any]] = []
    for path in paths:
        if not os.path.isfile(path):
            continue
        process_yaml_file(errors, path, yamllint_config)

    errors.sort(
        key=lambda error: (error["path"], error["line"], error["col"], error["message"])
    )
    for error in errors:
        prefix = f"{error['path']}:{error['line']}:{error['col']}: "
        msg = error["message"]
        if "note" in error:
            msg = f"{msg}\nNote: {error['note']}"
        for i, line in enumerate(msg.splitlines()):
            print(f"{prefix}{line}")
            if i == 0:
                prefix = " " * len(prefix)

    return len(errors) > 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

# Copyright (c) 2025, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Retrieve the version of one or more packages."""

from __future__ import annotations

import json
import sys
from importlib.metadata import PackageNotFoundError, version

from antsibull_nox_data_util import get_list_of_strings, setup  # type: ignore


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()

    packages = get_list_of_strings(extra_data, "packages", default=[])

    result: dict[str, str | None] = {}
    for package in packages:
        try:
            result[package] = version(package)
        except PackageNotFoundError:
            result[package] = None
    print(json.dumps(result))
    return 0


if __name__ == "__main__":
    sys.exit(main())



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Ansible-core version utilities.
"""

from __future__ import annotations

import functools
import typing as t
from dataclasses import dataclass

from antsibull_fileutils.yaml import load_yaml_file
from packaging.specifiers import SpecifierSet as PckgSpecifierSet
from packaging.version import Version as PckgVersion

from .utils import Version, version_range

AnsibleCoreVersion = t.Union[Version, t.Literal["milestone", "devel"]]


@dataclass(frozen=True)
class AnsibleCoreInfo:
    """
    Information on an ansible-core version.
    """

    ansible_core_version: Version
    controller_python_versions: tuple[Version, ...]
    remote_python_versions: tuple[Version, ...]


_MIN_SUPPORTED_VERSION = Version.parse("2.9")
_CURRENT_DEVEL_VERSION = Version.parse("2.20")
_CURRENT_MILESTONE_VERSION = Version.parse("2.20")

_SUPPORTED_CORE_VERSIONS: dict[Version | t.Literal["milestone"], AnsibleCoreInfo] = {
    (
        "milestone"
        if ansible_version == "milestone"
        else Version.parse(ansible_version)
    ): AnsibleCoreInfo(
        ansible_core_version=(
            _CURRENT_MILESTONE_VERSION
            if ansible_version == "milestone"
            else Version.parse(ansible_version)
        ),
        controller_python_versions=tuple(
            Version.parse(v) for v in controller_python_versions
        ),
        remote_python_versions=tuple(Version.parse(v) for v in remote_python_versions),
    )
    for ansible_version, (controller_python_versions, remote_python_versions) in {
        "2.9": [
            ["2.7", "3.5", "3.6", "3.7", "3.8"],
            ["2.6", "2.7", "3.5", "3.6", "3.7", "3.8"],
        ],
        "2.10": [
            ["2.7", "3.5", "3.6", "3.7", "3.8", "3.9"],
            ["2.6", "2.7", "3.5", "3.6", "3.7", "3.8", "3.9"],
        ],
        "2.11": [
            ["2.7", "3.5", "3.6", "3.7", "3.8", "3.9"],
            ["2.6", "2.7", "3.5", "3.6", "3.7", "3.8", "3.9"],
        ],
        "2.12": [
            ["3.8", "3.9", "3.10"],
            ["2.6", "2.7", "3.5", "3.6", "3.7", "3.8", "3.9", "3.10"],
        ],
        "2.13": [
            ["3.8", "3.9", "3.10"],
            ["2.7", "3.5", "3.6", "3.7", "3.8", "3.9", "3.10"],
        ],
        "2.14": [
            ["3.9", "3.10", "3.11"],
            ["2.7", "3.5", "3.6", "3.7", "3.8", "3.9", "3.10", "3.11"],
        ],
        "2.15": [
            ["3.9", "3.10", "3.11"],
            ["2.7", "3.5", "3.6", "3.7", "3.8", "3.9", "3.10", "3.11"],
        ],
        "2.16": [
            ["3.10", "3.11", "3.12"],
            ["2.7", "3.6", "3.7", "3.8", "3.9", "3.10", "3.11", "3.12"],
        ],
        "2.17": [
            ["3.10", "3.11", "3.12"],
            ["3.7", "3.8", "3.9", "3.10", "3.11", "3.12"],
        ],
        "2.18": [
            ["3.11", "3.12", "3.13"],
            ["3.8", "3.9", "3.10", "3.11", "3.12", "3.13"],
        ],
        "2.19": [
            ["3.11", "3.12", "3.13"],
            ["3.8", "3.9", "3.10", "3.11", "3.12", "3.13"],
        ],
        "2.20": [
            ["3.11", "3.12", "3.13"],
            ["3.9", "3.10", "3.11", "3.12", "3.13"],
            # Note: 3.14 should get added, 3.11 removed
        ],
        "milestone": [
            ["3.11", "3.12", "3.13"],
            ["3.9", "3.10", "3.11", "3.12", "3.13"],
            # Note: might lag behind devel
            # Note: 3.14 should get added, 3.11 removed
        ],
        # The following might need updates. Look for the "``ansible-core`` support matrix" table in:
        # https://github.com/ansible/ansible-documentation/blob/devel/docs/docsite/rst/reference_appendices/release_and_maintenance.rst?plain=1
        # It contains commented-out entries for future ansible-core versions.
        "2.21": [
            ["3.12", "3.13", "3.14"],
            ["3.9", "3.10", "3.11", "3.12", "3.13", "3.14"],
        ],
        "2.22": [
            ["3.13", "3.14", "3.15"],
            ["3.10", "3.11", "3.12", "3.13", "3.14", "3.15"],
        ],
        "2.23": [
            ["3.13", "3.14", "3.15"],
            ["3.10", "3.11", "3.12", "3.13", "3.14", "3.15"],
        ],
        "2.24": [
            ["3.14", "3.15", "3.16"],
            ["3.11", "3.12", "3.13", "3.14", "3.15", "3.16"],
        ],
        "2.25": [
            ["3.14", "3.15", "3.16"],
            ["3.11", "3.12", "3.13", "3.14", "3.15", "3.16"],
        ],
    }.items()
}


def get_ansible_core_info(
    core_version: AnsibleCoreVersion,
) -> AnsibleCoreInfo:
    """
    Retrieve information on an ansible-core version.
    """
    version: Version
    if core_version == "devel":
        version = _CURRENT_DEVEL_VERSION
    elif core_version == "milestone":
        version = _CURRENT_MILESTONE_VERSION
    else:
        version = core_version
    if version in _SUPPORTED_CORE_VERSIONS:
        return _SUPPORTED_CORE_VERSIONS[version]
    raise ValueError(f"Unknown ansible-core version {version}")


_ANSIBLE_REPO = "ansible/ansible"
_ANSIBLE_EOL_REPO = "ansible-community/eol-ansible"
_ANSIBLE_EOL_MAX_VERSION = Version(2, 14)


def get_ansible_core_package_name(
    core_version: AnsibleCoreVersion,
    *,
    source: t.Literal["git", "pypi"] = "git",
    ansible_repo: str | None = None,
    branch_name: str | None = None,
) -> str:
    """
    Return the package name for a specific ansible-core version.

    The result can be passed to pip to install that version of ansible-core.
    """
    if not isinstance(core_version, Version):
        # milestone and devel are not available from PyPI
        source = "git"

    if source == "git":
        if branch_name is None:
            if isinstance(core_version, str):
                branch_name = core_version
            else:
                branch_name = f"stable-{core_version}"
        if ansible_repo is None:
            if (
                isinstance(core_version, Version)
                and core_version <= _ANSIBLE_EOL_MAX_VERSION
            ):
                ansible_repo = _ANSIBLE_EOL_REPO
            else:
                ansible_repo = _ANSIBLE_REPO
        return f"https://github.com/{ansible_repo}/archive/{branch_name}.tar.gz"

    assert isinstance(core_version, Version)
    next_core_version = core_version.next_minor_version()
    base = "ansible-core"
    if core_version == Version(2, 9):
        base = "ansible"
    elif core_version == Version(2, 10):
        base = "ansible-base"
    return f"{base}>={core_version},<{next_core_version}"


@functools.cache
def _read_requires_ansible() -> PckgSpecifierSet:
    path = "meta/runtime.yml"
    try:
        runtime_data = load_yaml_file(path)
    except FileNotFoundError as exc:
        raise ValueError(f"Cannot open {path}") from exc
    except Exception as exc:
        raise ValueError(f"Cannot parse {path}: {exc}") from exc

    requires_ansible = runtime_data.get("requires_ansible")
    if not isinstance(requires_ansible, str):
        raise ValueError(f"{path} does not contain an 'requires_ansible' string")
    try:
        return PckgSpecifierSet(requires_ansible)
    except Exception as exc:
        raise ValueError(
            f"{path} contains an invalid 'requires_ansible' string: {exc}"
        ) from exc


@functools.cache
def get_supported_core_versions(
    *,
    include_devel: bool = False,
    include_milestone: bool = False,
    min_version: Version | None = None,
    max_version: Version | None = None,
    except_versions: tuple[AnsibleCoreVersion, ...] | None = None,
) -> list[AnsibleCoreVersion]:
    """
    Extracts a list of supported ansible-core versions from meta/runtime.yml.

    If ``min_version`` is specified, no version below that version will be returned.
    If ``max_version`` is specified, no version above that version will be returned.
    If ``except_versions`` is specified, no version in that tuple will be returned.
    """
    if except_versions is None:
        except_versions = ()

    ra_specifier = _read_requires_ansible()

    result: list[AnsibleCoreVersion] = []
    for version in version_range(
        _MIN_SUPPORTED_VERSION, _CURRENT_DEVEL_VERSION, inclusive=False
    ):
        if version in except_versions:
            continue
        if min_version is not None and version < min_version:
            continue
        if max_version is not None and version > max_version:
            continue
        # We're using x.y.999 to check whether *some* ansible-core x.y version is supported.
        # This is not entirely correct, since collections might specfiy that only certain older x.y
        # versions are OK, but I'd consider such behavior a bug of the collection and something
        # you really shouldn't do as a collection maintainer.
        v = PckgVersion(f"{version.major}.{version.minor}.999")
        if v in ra_specifier:
            result.append(version)
    if include_milestone and "milestone" not in except_versions:
        result.append("milestone")
    if include_devel and "devel" not in except_versions:
        result.append("devel")
    return result


def parse_ansible_core_version(
    version: str | AnsibleCoreVersion,
) -> AnsibleCoreVersion:
    """
    Coerce a string or a AnsibleCoreVersion to a AnsibleCoreVersion.
    """
    if version in ("devel", "milestone"):
        # For some reason mypy doesn't notice that
        return t.cast(AnsibleCoreVersion, version)
    if isinstance(version, Version):
        return version
    return Version.parse(version)


__all__ = [
    "AnsibleCoreInfo",
    "get_ansible_core_info",
    "get_ansible_core_package_name",
    "parse_ansible_core_version",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Build Ansible collections.
"""

from __future__ import annotations

from pathlib import Path

from antsibull_fileutils.yaml import load_yaml_file, store_yaml_file

from .search import GALAXY_YML


def force_collection_version(path: Path, *, version: str) -> bool:
    """
    Make sure galaxy.yml contains this version.

    Returns ``True`` if the version was changed, and ``False`` if the version
    was already set to this value.
    """
    galaxy_yml = path / GALAXY_YML
    try:
        data = load_yaml_file(galaxy_yml)
    except Exception as exc:
        raise ValueError(f"Cannot parse {galaxy_yml}: {exc}") from exc
    if data.get("version") == version:
        return False
    data["version"] = version
    store_yaml_file(galaxy_yml, data)
    return True


__all__ = [
    "force_collection_version",
]



#!/usr/bin/env python

# Copyright (c) 2025, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Run antsibull-nox lint-config."""

from __future__ import annotations

import sys

from antsibull_nox.data.antsibull_nox_data_util import setup
from antsibull_nox.lint_config import lint_config


def main() -> int:
    """Main entry point."""
    _, __ = setup()

    errors = lint_config()
    for error in errors:
        print(error)
    return len(errors) > 0


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

# Copyright (c) 2025, Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt
# or https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later

"""Avoid characters / regexes in files."""

from __future__ import annotations

import os
import re
import sys

from antsibull_nox.data.antsibull_nox_data_util import setup
from antsibull_nox.sessions.extra_checks import AvoidCharacterGroup


def compile_patterns(
    config: list[AvoidCharacterGroup], errors: list[str]
) -> list[re.Pattern]:
    patterns: list[re.Pattern] = []
    for group in config:
        patterns.append(re.compile(group.regex))
    return patterns


class File:
    def __init__(self, path: str) -> None:
        self.path = path
        self.content: str | None = None

    def get_content(self) -> str:
        if self.content is None:
            with open(self.path, "r", encoding="utf-8") as f:
                self.content = f.read()
        return self.content


def get_position(content: str, position: int) -> tuple[int, int]:
    try:
        prev_newline_index = content.rindex("\n", 0, position + 1)
        return (
            2 + content[0:prev_newline_index].count("\n"),
            position - prev_newline_index,
        )
    except ValueError:
        # No previous newline, i.e. this is in the first line
        return 1, position + 1


def check(
    file: File, group: AvoidCharacterGroup, pattern: re.Pattern, errors: list[str]
) -> None:
    # Check whether the file is included
    if group.match_extensions is not None:
        if not any(
            file.path.endswith(extension) for extension in group.match_extensions
        ):
            return
    if group.match_paths is not None:
        if file.path not in group.match_paths:
            return
    if group.match_directories is not None:
        if not any(
            file.path.startswith(skip_directory)
            for skip_directory in group.match_directories
        ):
            return

    # Check whether the file is excluded
    if any(file.path.endswith(extension) for extension in group.skip_extensions):
        return
    if file.path in group.skip_paths:
        return
    if any(
        file.path.startswith(skip_directory)
        for skip_directory in group.skip_directories
    ):
        return

    content = file.get_content()
    m = pattern.search(content)
    if not m:
        return

    name = group.name if group.name is not None else group.regex
    error = f"found {name}"

    line, col = get_position(content, m.start())

    errors.append(f"{file.path}:{line}:{col}: {error}")


def scan(
    paths: list[str], config: list[AvoidCharacterGroup], errors: list[str]
) -> None:
    patterns = compile_patterns(config, errors)

    for path in paths:
        if not os.path.isfile(path):
            continue

        try:
            file = File(path)
            for index, group in enumerate(config):
                check(file, group, patterns[index], errors)
        except UnicodeDecodeError:
            errors.append(f"{path}: cannot parse file as UTF-8")
        except Exception as e:
            errors.append(f"{path}: unexpected error: {e}")


def main() -> int:
    """Main entry point."""
    paths, extra_data = setup()

    if not isinstance(extra_data.get("config"), list):
        raise ValueError("config is not a list")
    if not all(isinstance(cfg, dict) for cfg in extra_data["config"]):
        raise ValueError("config is not a list of dictionaries")
    config = [AvoidCharacterGroup(**cfg) for cfg in extra_data["config"]]

    errors: list[str] = []
    scan(paths, config, errors)

    for error in sorted(errors):
        print(error)
    return len(errors) > 0


if __name__ == "__main__":
    sys.exit(main())



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox sessions.
"""

from __future__ import annotations

import os
import subprocess
import typing as t
from dataclasses import dataclass
from pathlib import Path

import nox

from ..ansible import AnsibleCoreVersion, parse_ansible_core_version
from ..collection import (
    CollectionData,
    setup_collections,
    setup_current_tree,
)
from ..paths import (
    create_temp_directory,
)


@dataclass
class CollectionSetup:
    """
    Information on the setup collections.
    """

    # The path of the ansible_collections directory where all dependent collections
    # are installed. Is currently identical to current_root, but that might change
    # or depend on options in the future.
    collections_root: Path

    # The directory in which ansible_collections can be found, as well as
    # ansible_collections/<namespace>/<name> points to a copy of the current collection.
    current_place: Path

    # The path of the ansible_collections directory that contains the current collection.
    # The following is always true:
    #   current_root == current_place / "ansible_collections"
    current_root: Path

    # Data on the current collection (as in the repository).
    current_collection: CollectionData

    # The path of the current collection inside the collection tree below current_root.
    # The following is always true:
    #   current_path == current_root / current_collection.namespace / current_collection.name
    current_path: Path

    def prefix_current_paths(self, paths: list[str]) -> list[str]:
        """
        Prefix the list of given paths with ``current_path``.
        """
        result = []
        for path in paths:
            prefixed_path = (self.current_path / path).relative_to(self.current_place)
            if prefixed_path.exists():
                result.append(str(prefixed_path))
        return result


def _run_subprocess(args: list[str]) -> tuple[bytes, bytes]:
    p = subprocess.run(args, check=True, capture_output=True)
    return p.stdout, p.stderr


def prepare_collections(
    session: nox.Session,
    *,
    ansible_core_version: AnsibleCoreVersion | str | None = None,
    install_in_site_packages: bool,
    extra_deps_files: list[str | os.PathLike] | None = None,
    extra_collections: list[str] | None = None,
    install_out_of_tree: bool = False,  # can not be used with install_in_site_packages=True
) -> CollectionSetup | None:
    """
    Install collections in site-packages.
    """
    parsed_ansible_core_version = (
        parse_ansible_core_version(ansible_core_version)
        if ansible_core_version is not None
        else "devel"
    )
    if install_out_of_tree and install_in_site_packages:
        raise ValueError(
            "install_out_of_tree=True cannot be combined with install_in_site_packages=True"
        )
    if isinstance(session.virtualenv, nox.virtualenv.PassthroughEnv):
        session.warn("No venv. Skip preparing collections...")
        return None
    if install_in_site_packages:
        purelib = (
            session.run(
                "python",
                "-c",
                "import sysconfig; print(sysconfig.get_path('purelib'))",
                silent=True,
            )
            or ""
        ).strip()
        if not purelib:
            session.warn(
                "Cannot find site-packages (probably due to install-only run)."
                " Skip preparing collections..."
            )
            return None
        place = Path(purelib)
    elif install_out_of_tree:
        place = create_temp_directory(f"antsibull-nox-{session.name}-collection-root-")
    else:
        place = Path(session.virtualenv.location) / "collection-root"
    place.mkdir(exist_ok=True)
    setup = setup_collections(
        place,
        _run_subprocess,
        ansible_core_version=parsed_ansible_core_version,
        extra_deps_files=extra_deps_files,
        extra_collections=extra_collections,
        with_current=False,
        global_cache_dir=session.cache_dir,
    )
    current_setup = setup_current_tree(place, setup.current_collection)
    return CollectionSetup(
        collections_root=setup.root,
        current_place=place,
        current_root=current_setup.root,
        current_collection=setup.current_collection,
        current_path=t.cast(Path, current_setup.current_path),
    )


__all__ = [
    "CollectionSetup",
    "prepare_collections",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox lint sessions.
"""

from __future__ import annotations

import json
import os
import shlex
from pathlib import Path

import nox

from ..paths import (
    filter_paths,
    list_all_files,
)
from .collections import (
    CollectionSetup,
    prepare_collections,
)
from .docs_check import find_extra_docs_rst_files
from .utils import (
    IN_CI,
    compose_description,
    install,
    run_bare_script,
    silence_run_verbosity,
)

CODE_FILES = [
    "plugins",
    "tests/unit",
]

MODULE_PATHS = [
    "plugins/modules/",
    "plugins/module_utils/",
    "tests/unit/plugins/modules/",
    "tests/unit/plugins/module_utils/",
]


def add_lint(
    *,
    make_lint_default: bool,
    has_formatters: bool,
    has_codeqa: bool,
    has_yamllint: bool,
    has_typing: bool,
    has_config_lint: bool,
) -> None:
    """
    Add nox meta session for linting.
    """

    def lint(session: nox.Session) -> None:  # pylint: disable=unused-argument
        pass  # this session is deliberately empty

    dependent_sessions = []
    if has_formatters:
        dependent_sessions.append("formatters")
    if has_codeqa:
        dependent_sessions.append("codeqa")
    if has_yamllint:
        dependent_sessions.append("yamllint")
    if has_typing:
        dependent_sessions.append("typing")
    if has_config_lint:
        dependent_sessions.append("antsibull-nox-config")

    lint.__doc__ = compose_description(
        prefix={
            "one": "Meta session for triggering the following session:",
            "other": "Meta session for triggering the following sessions:",
        },
        programs={
            "formatters": has_formatters,
            "codeqa": has_codeqa,
            "yamllint": has_yamllint,
            "typing": has_typing,
            "antsibull-nox-config": has_config_lint,
        },
    )
    nox.session(
        name="lint",
        default=make_lint_default,
        requires=dependent_sessions,
    )(lint)


def _execute_isort(
    session: nox.Session,
    *,
    run_check: bool,
    extra_code_files: list[str],
    isort_config: str | os.PathLike | None,
) -> None:
    command: list[str] = [
        "isort",
    ]
    if run_check:
        command.append("--check")
    if isort_config is not None:
        command.extend(["--settings-file", str(isort_config)])
    command.extend(session.posargs)
    command.extend(filter_paths(CODE_FILES + ["noxfile.py"] + extra_code_files))
    session.run(*command)


def _execute_black_for(
    session: nox.Session,
    *,
    paths: list[str],
    run_check: bool,
    black_config: str | os.PathLike | None,
) -> None:
    if not paths:
        return
    command = ["black"]
    if run_check:
        command.append("--check")
    if black_config is not None:
        command.extend(["--config", str(black_config)])
    command.extend(session.posargs)
    command.extend(paths)
    session.run(*command)


def _execute_black(
    session: nox.Session,
    *,
    run_check: bool,
    extra_code_files: list[str],
    run_black: bool,
    run_black_modules: bool | None,
    black_config: str | os.PathLike | None,
) -> None:
    if run_black and run_black_modules:
        _execute_black_for(
            session,
            paths=filter_paths(CODE_FILES + ["noxfile.py"] + extra_code_files),
            run_check=run_check,
            black_config=black_config,
        )
        return
    if run_black:
        paths = (
            filter_paths(
                CODE_FILES,
                remove=MODULE_PATHS,
                extensions=[".py"],
            )
            + ["noxfile.py"]
            + extra_code_files
        )
        _execute_black_for(
            session, paths=paths, run_check=run_check, black_config=black_config
        )
    if run_black_modules:
        paths = filter_paths(
            CODE_FILES,
            restrict=MODULE_PATHS,
            extensions=[".py"],
        )
        _execute_black_for(
            session, paths=paths, run_check=run_check, black_config=black_config
        )


def _execute_ruff_format(
    session: nox.Session,
    *,
    run_check: bool,
    extra_code_files: list[str],
    ruff_format_config: str | os.PathLike | None,
) -> None:
    command: list[str] = [
        "ruff",
        "format",
    ]
    if run_check:
        command.append("--check")
    if ruff_format_config is not None:
        command.extend(["--config", str(ruff_format_config)])
    command.extend(session.posargs)
    command.extend(filter_paths(CODE_FILES + ["noxfile.py"] + extra_code_files))
    session.run(*command)


def _execute_ruff_autofix(
    session: nox.Session,
    *,
    run_check: bool,
    extra_code_files: list[str],
    ruff_autofix_config: str | os.PathLike | None,
    ruff_autofix_select: list[str],
) -> None:
    command: list[str] = [
        "ruff",
        "check",
    ]
    if not run_check:
        command.append("--fix")
    if ruff_autofix_config is not None:
        command.extend(["--config", str(ruff_autofix_config)])
    if ruff_autofix_select:
        command.extend(["--select", ",".join(ruff_autofix_select)])
    command.extend(session.posargs)
    command.extend(filter_paths(CODE_FILES + ["noxfile.py"] + extra_code_files))
    session.run(*command)


def add_formatters(
    *,
    extra_code_files: list[str],
    # isort:
    run_isort: bool,
    isort_config: str | os.PathLike | None,
    isort_package: str,
    # black:
    run_black: bool,
    run_black_modules: bool | None,
    black_config: str | os.PathLike | None,
    black_package: str,
    # ruff format:
    run_ruff_format: bool,
    ruff_format_config: str | os.PathLike | None,
    ruff_format_package: str,
    # ruff autofix:
    run_ruff_autofix: bool,
    ruff_autofix_config: str | os.PathLike | None,
    ruff_autofix_package: str,
    ruff_autofix_select: list[str],
) -> None:
    """
    Add nox session for formatters.
    """
    if run_black_modules is None:
        run_black_modules = run_black
    run_check = IN_CI

    def compose_dependencies() -> list[str]:
        deps = []
        if run_isort:
            deps.append(isort_package)
        if run_black or run_black_modules:
            deps.append(black_package)
        if (
            run_ruff_format
            and run_ruff_autofix
            and ruff_format_package == ruff_autofix_package
        ):
            deps.append(ruff_format_package)
        else:
            if run_ruff_format:
                deps.append(ruff_format_package)
            if run_ruff_autofix:
                deps.append(ruff_autofix_package)
        return deps

    def formatters(session: nox.Session) -> None:
        install(session, *compose_dependencies())
        if run_isort:
            _execute_isort(
                session,
                run_check=run_check,
                extra_code_files=extra_code_files,
                isort_config=isort_config,
            )
        if run_black or run_black_modules:
            _execute_black(
                session,
                run_check=run_check,
                extra_code_files=extra_code_files,
                run_black=run_black,
                run_black_modules=run_black_modules,
                black_config=black_config,
            )
        if run_ruff_format:
            _execute_ruff_format(
                session,
                run_check=run_check,
                extra_code_files=extra_code_files,
                ruff_format_config=ruff_format_config,
            )
        if run_ruff_autofix:
            _execute_ruff_autofix(
                session,
                run_check=run_check,
                extra_code_files=extra_code_files,
                ruff_autofix_config=ruff_autofix_config,
                ruff_autofix_select=ruff_autofix_select,
            )

    formatters.__doc__ = compose_description(
        prefix={
            "one": "Run code formatter:",
            "other": "Run code formatters:",
        },
        programs={
            "isort": run_isort,
            "black": run_black,
            "ruff format": run_ruff_format,
            "ruff check --fix": run_ruff_autofix,
        },
    )
    nox.session(name="formatters", default=False)(formatters)


def process_pylint_errors(
    session: nox.Session,
    prepared_collections: CollectionSetup,
    output: str,
) -> None:
    """
    Process errors reported by pylint in 'json2' format.
    """
    data = json.loads(output)
    found_error = False
    if data["messages"]:
        for message in data["messages"]:
            path = os.path.relpath(
                message["absolutePath"], prepared_collections.current_path
            )
            prefix = f"{path}:{message['line']}:{message['column']}: [{message['messageId']}]"
            print(f"{prefix} {message['message']} [{message['symbol']}]")
            found_error = True
    if found_error:
        session.error("Pylint failed")


def add_codeqa(  # noqa: C901
    *,
    extra_code_files: list[str],
    # ruff check:
    run_ruff_check: bool,
    ruff_check_config: str | os.PathLike | None,
    ruff_check_package: str,
    # flake8:
    run_flake8: bool,
    flake8_config: str | os.PathLike | None,
    flake8_package: str,
    # pylint:
    run_pylint: bool,
    pylint_rcfile: str | os.PathLike | None,
    pylint_modules_rcfile: str | os.PathLike | None,
    pylint_package: str,
    pylint_ansible_core_package: str | None,
    pylint_extra_deps: list[str],
) -> None:
    """
    Add nox session for codeqa.
    """

    def compose_dependencies() -> list[str]:
        deps = []
        if run_ruff_check:
            deps.append(ruff_check_package)
        if run_flake8:
            deps.append(flake8_package)
        if run_pylint:
            deps.append(pylint_package)
            if pylint_ansible_core_package is not None:
                deps.append(pylint_ansible_core_package)
            if os.path.isdir("tests/unit"):
                deps.append("pytest")
                if os.path.isfile("tests/unit/requirements.txt"):
                    deps.extend(["-r", "tests/unit/requirements.txt"])
            for extra_dep in pylint_extra_deps:
                deps.extend(shlex.split(extra_dep))
        return deps

    def execute_ruff_check(session: nox.Session) -> None:
        command: list[str] = [
            "ruff",
            "check",
        ]
        if ruff_check_config is not None:
            command.extend(["--config", str(ruff_check_config)])
        command.extend(session.posargs)
        command.extend(filter_paths(CODE_FILES + ["noxfile.py"] + extra_code_files))
        session.run(*command)

    def execute_flake8(session: nox.Session) -> None:
        command: list[str] = [
            "flake8",
        ]
        if flake8_config is not None:
            command.extend(["--config", str(flake8_config)])
        command.extend(session.posargs)
        command.extend(filter_paths(CODE_FILES + ["noxfile.py"] + extra_code_files))
        session.run(*command)

    def execute_pylint_impl(
        session: nox.Session,
        prepared_collections: CollectionSetup,
        config: os.PathLike | str | None,
        paths: list[str],
    ) -> None:
        command = ["pylint"]
        if config is not None:
            command.extend(
                [
                    "--rcfile",
                    os.path.join(prepared_collections.current_collection.path, config),
                ]
            )
        command.extend(["--source-roots", "."])
        command.extend(["--output-format", "json2"])
        command.extend(session.posargs)
        command.extend(prepared_collections.prefix_current_paths(paths))
        with silence_run_verbosity():
            # Exit code is OR of some of 1, 2, 4, 8, 16
            output = session.run(
                *command, silent=True, success_codes=list(range(0, 32))
            )

        if output:
            process_pylint_errors(session, prepared_collections, output)

    def execute_pylint(
        session: nox.Session, prepared_collections: CollectionSetup
    ) -> None:
        if pylint_modules_rcfile is not None and pylint_modules_rcfile != pylint_rcfile:
            # Only run pylint twice when using different configurations
            module_paths = filter_paths(
                CODE_FILES, restrict=MODULE_PATHS, extensions=[".py"]
            )
            other_paths = filter_paths(
                CODE_FILES, remove=MODULE_PATHS, extensions=[".py"]
            )
        else:
            # Otherwise run it only once using the general configuration
            module_paths = []
            other_paths = filter_paths(CODE_FILES)

        with session.chdir(prepared_collections.current_place):
            if module_paths:
                execute_pylint_impl(
                    session,
                    prepared_collections,
                    pylint_modules_rcfile or pylint_rcfile,
                    module_paths,
                )

            if other_paths:
                execute_pylint_impl(
                    session, prepared_collections, pylint_rcfile, other_paths
                )

    def codeqa(session: nox.Session) -> None:
        install(session, *compose_dependencies())
        prepared_collections: CollectionSetup | None = None
        if run_pylint:
            prepared_collections = prepare_collections(
                session,
                install_in_site_packages=False,
                extra_deps_files=["tests/unit/requirements.yml"],
            )
            if not prepared_collections:
                session.warn("Skipping pylint...")
        if run_ruff_check:
            execute_ruff_check(session)
        if run_flake8:
            execute_flake8(session)
        if run_pylint and prepared_collections:
            execute_pylint(session, prepared_collections)

    codeqa.__doc__ = compose_description(
        prefix={
            "other": "Run code QA:",
        },
        programs={
            "ruff check": run_ruff_check,
            "flake8": run_flake8,
            "pylint": run_pylint,
        },
    )
    nox.session(name="codeqa", default=False)(codeqa)


def add_yamllint(
    *,
    run_yamllint: bool,
    yamllint_config: str | os.PathLike | None,
    yamllint_config_plugins: str | os.PathLike | None,
    yamllint_config_plugins_examples: str | os.PathLike | None,
    yamllint_config_extra_docs: str | os.PathLike | None,
    yamllint_package: str,
    yamllint_antsibull_docutils_package: str,
) -> None:
    """
    Add yamllint session for linting YAML files and plugin/module docs.
    """

    def compose_dependencies() -> list[str]:
        deps = []
        if run_yamllint:
            deps.extend([yamllint_package, yamllint_antsibull_docutils_package])
        return deps

    def to_str(config: str | os.PathLike | None) -> str | None:
        return str(config) if config else None

    def execute_yamllint(session: nox.Session) -> None:
        all_files = list_all_files()
        all_yaml_filenames = [
            file for file in all_files if file.name.lower().endswith((".yml", ".yaml"))
        ]
        if not all_yaml_filenames:
            session.warn("Skipping yamllint since no YAML file was found...")
            return

        run_bare_script(
            session,
            "file-yamllint",
            use_session_python=True,
            files=all_yaml_filenames,
            extra_data={
                "config": to_str(yamllint_config),
            },
        )

    def execute_plugin_yamllint(session: nox.Session) -> None:
        all_files = list_all_files()
        cwd = Path.cwd()
        plugins_dir = cwd / "plugins"
        ignore_dirs = [
            plugins_dir / "action",
            plugins_dir / "module_utils",
            plugins_dir / "plugin_utils",
        ]
        all_plugin_files = [
            file
            for file in all_files
            if file.is_relative_to(plugins_dir)
            and file.name.lower().endswith((".py", ".yml", ".yaml"))
            and not any(file.is_relative_to(dir) for dir in ignore_dirs)
        ]
        if not all_plugin_files:
            session.warn(
                "Skipping yamllint for modules/plugins since"
                " no appropriate Python file was found..."
            )
            return
        run_bare_script(
            session,
            "plugin-yamllint",
            use_session_python=True,
            files=all_plugin_files,
            extra_data={
                "config": to_str(yamllint_config_plugins or yamllint_config),
                "config_examples": to_str(
                    yamllint_config_plugins_examples
                    or yamllint_config_plugins
                    or yamllint_config
                ),
            },
        )

    def execute_extra_docs_yamllint(session: nox.Session) -> None:
        all_extra_docs = find_extra_docs_rst_files()
        if not all_extra_docs:
            session.warn(
                "Skipping yamllint for extra docs since"
                " no appropriate RST file was found..."
            )
            return
        run_bare_script(
            session,
            "rst-yamllint",
            use_session_python=True,
            files=all_extra_docs,
            extra_data={
                "config": to_str(
                    yamllint_config_extra_docs
                    or yamllint_config_plugins_examples
                    or yamllint_config_plugins
                    or yamllint_config
                ),
            },
        )

    def yamllint(session: nox.Session) -> None:
        install(session, *compose_dependencies())
        if run_yamllint:
            execute_yamllint(session)
            execute_plugin_yamllint(session)
            execute_extra_docs_yamllint(session)

    yamllint.__doc__ = compose_description(
        prefix={
            "one": "Run YAML checker:",
            "other": "Run YAML checkers:",
        },
        programs={
            "yamllint": run_yamllint,
        },
    )
    nox.session(name="yamllint", default=False)(yamllint)


def process_mypy_errors(
    session: nox.Session,
    prepared_collections: CollectionSetup,
    output: str,
) -> None:
    """
    Process errors reported by mypy in 'json' format.
    """
    found_error = False
    for line in output.splitlines():
        if not line.strip():
            continue
        try:
            data = json.loads(line)
            path = os.path.relpath(
                prepared_collections.current_place / data["file"],
                prepared_collections.current_path,
            )
            prefix = f"{path}:{data['line']}:{data['column']}: [{data['severity']}]"
            if data["code"]:
                print(f"{prefix} {data['message']} [{data['code']}]")
            else:
                print(f"{prefix} {data['message']}")
            if data["hint"]:
                prefix = " " * len(prefix)
                for hint in data["hint"].splitlines():
                    print(f"{prefix} {hint}")
        except Exception:  # pylint: disable=broad-exception-caught
            session.warn(f"Cannot parse mypy output: {line}")
        found_error = True
    if found_error:
        session.error("Type checking failed")


def add_typing(
    *,
    extra_code_files: list[str],
    run_mypy: bool,
    mypy_config: str | os.PathLike | None,
    mypy_package: str,
    mypy_ansible_core_package: str | None,
    mypy_extra_deps: list[str],
) -> None:
    """
    Add nox session for typing.
    """

    def compose_dependencies() -> list[str]:
        deps = []
        if run_mypy:
            deps.append(mypy_package)
            if mypy_ansible_core_package is not None:
                deps.append(mypy_ansible_core_package)
            if os.path.isdir("tests/unit"):
                deps.append("pytest")
                if os.path.isfile("tests/unit/requirements.txt"):
                    deps.extend(["-r", "tests/unit/requirements.txt"])
            for extra_dep in mypy_extra_deps:
                deps.extend(shlex.split(extra_dep))
        return deps

    def execute_mypy(
        session: nox.Session, prepared_collections: CollectionSetup
    ) -> None:
        # Run mypy
        with session.chdir(prepared_collections.current_place):
            command = ["mypy"]
            if mypy_config is not None:
                command.extend(
                    [
                        "--config-file",
                        os.path.join(
                            prepared_collections.current_collection.path, mypy_config
                        ),
                    ]
                )
            command.append("--namespace-packages")
            command.append("--explicit-package-bases")
            command.extend(["--output", "json"])
            command.extend(session.posargs)
            command.extend(
                prepared_collections.prefix_current_paths(CODE_FILES + extra_code_files)
            )
            with silence_run_verbosity():
                output = session.run(
                    *command,
                    env={"MYPYPATH": str(prepared_collections.current_place)},
                    silent=True,
                    success_codes=(0, 1, 2),
                )

            if output:
                process_mypy_errors(session, prepared_collections, output)

    def typing(session: nox.Session) -> None:
        install(session, *compose_dependencies())
        prepared_collections = prepare_collections(
            session,
            install_in_site_packages=False,
            extra_deps_files=["tests/unit/requirements.yml"],
        )
        if not prepared_collections:
            session.warn("Skipping mypy...")
        if run_mypy and prepared_collections:
            execute_mypy(session, prepared_collections)

    typing.__doc__ = compose_description(
        prefix={
            "one": "Run type checker:",
            "other": "Run type checkers:",
        },
        programs={
            "mypy": run_mypy,
        },
    )
    nox.session(name="typing", default=False)(typing)


def add_config_lint(
    *,
    run_antsibullnox_config_lint: bool,
):
    """
    Add nox session for antsibull-nox config linting.
    """

    def antsibull_nox_config(session: nox.Session) -> None:
        if run_antsibullnox_config_lint:
            run_bare_script(
                session,
                "antsibull-nox-lint-config",
            )

            session.run("antsibull-nox", "lint-config")

    antsibull_nox_config.__doc__ = "Lint antsibull-nox config"
    nox.session(name="antsibull-nox-config", python=False, default=False)(
        antsibull_nox_config
    )


def add_lint_sessions(
    *,
    make_lint_default: bool = True,
    extra_code_files: list[str] | None = None,
    # isort:
    run_isort: bool = True,
    isort_config: str | os.PathLike | None = None,
    isort_package: str = "isort",
    # black:
    run_black: bool = True,
    run_black_modules: bool | None = None,
    black_config: str | os.PathLike | None = None,
    black_package: str = "black",
    # ruff format:
    run_ruff_format: bool = False,
    ruff_format_config: str | os.PathLike | None = None,
    ruff_format_package: str = "ruff",
    # ruff autofix:
    run_ruff_autofix: bool = False,
    ruff_autofix_config: str | os.PathLike | None = None,
    ruff_autofix_package: str = "ruff",
    ruff_autofix_select: list[str] | None = None,
    # ruff check:
    run_ruff_check: bool = False,
    ruff_check_config: str | os.PathLike | None = None,
    ruff_check_package: str = "ruff",
    # flake8:
    run_flake8: bool = True,
    flake8_config: str | os.PathLike | None = None,
    flake8_package: str = "flake8",
    # pylint:
    run_pylint: bool = True,
    pylint_rcfile: str | os.PathLike | None = None,
    pylint_modules_rcfile: str | os.PathLike | None = None,
    pylint_package: str = "pylint",
    pylint_ansible_core_package: str | None = "ansible-core",
    pylint_extra_deps: list[str] | None = None,
    # yamllint:
    run_yamllint: bool = False,
    yamllint_config: str | os.PathLike | None = None,
    yamllint_config_plugins: str | os.PathLike | None = None,
    yamllint_config_plugins_examples: str | os.PathLike | None = None,
    yamllint_config_extra_docs: str | os.PathLike | None = None,
    yamllint_package: str = "yamllint",
    yamllint_antsibull_docutils_package: str = "antsibull-docutils",
    # mypy:
    run_mypy: bool = True,
    mypy_config: str | os.PathLike | None = None,
    mypy_package: str = "mypy",
    mypy_ansible_core_package: str | None = "ansible-core",
    mypy_extra_deps: list[str] | None = None,
    # antsibull-nox config lint:
    run_antsibullnox_config_lint: bool = True,
) -> None:
    """
    Add nox sessions for linting.
    """
    has_formatters = (
        run_isort
        or run_black
        or run_black_modules
        or False
        or run_ruff_format
        or run_ruff_autofix
    )
    has_codeqa = run_ruff_check or run_flake8 or run_pylint
    has_yamllint = run_yamllint
    has_typing = run_mypy
    has_config_lint = run_antsibullnox_config_lint

    add_lint(
        has_formatters=has_formatters,
        has_codeqa=has_codeqa,
        has_yamllint=has_yamllint,
        has_typing=has_typing,
        has_config_lint=has_config_lint,
        make_lint_default=make_lint_default,
    )

    if has_formatters:
        add_formatters(
            extra_code_files=extra_code_files or [],
            run_isort=run_isort,
            isort_config=isort_config,
            isort_package=isort_package,
            run_black=run_black,
            run_black_modules=run_black_modules,
            black_config=black_config,
            black_package=black_package,
            run_ruff_format=run_ruff_format,
            ruff_format_config=ruff_format_config,
            ruff_format_package=ruff_format_package,
            run_ruff_autofix=run_ruff_autofix,
            ruff_autofix_config=ruff_autofix_config,
            ruff_autofix_package=ruff_autofix_package,
            ruff_autofix_select=ruff_autofix_select or [],
        )

    if has_codeqa:
        add_codeqa(
            extra_code_files=extra_code_files or [],
            run_ruff_check=run_ruff_check,
            ruff_check_config=ruff_check_config,
            ruff_check_package=ruff_check_package,
            run_flake8=run_flake8,
            flake8_config=flake8_config,
            flake8_package=flake8_package,
            run_pylint=run_pylint,
            pylint_rcfile=pylint_rcfile,
            pylint_modules_rcfile=pylint_modules_rcfile,
            pylint_package=pylint_package,
            pylint_ansible_core_package=pylint_ansible_core_package,
            pylint_extra_deps=pylint_extra_deps or [],
        )

    if has_yamllint:
        add_yamllint(
            run_yamllint=run_yamllint,
            yamllint_config=yamllint_config,
            yamllint_config_plugins=yamllint_config_plugins,
            yamllint_config_plugins_examples=yamllint_config_plugins_examples,
            yamllint_config_extra_docs=yamllint_config_extra_docs,
            yamllint_package=yamllint_package,
            yamllint_antsibull_docutils_package=yamllint_antsibull_docutils_package,
        )

    if has_typing:
        add_typing(
            extra_code_files=extra_code_files or [],
            run_mypy=run_mypy,
            mypy_config=mypy_config,
            mypy_package=mypy_package,
            mypy_ansible_core_package=mypy_ansible_core_package,
            mypy_extra_deps=mypy_extra_deps or [],
        )

    if has_config_lint:
        add_config_lint(
            run_antsibullnox_config_lint=run_antsibullnox_config_lint,
        )


__all__ = [
    "add_lint_sessions",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Extract tarballs.
"""

from __future__ import annotations

import tarfile
from pathlib import Path


def extract_tarball(*, tarball: Path, destination: Path) -> None:
    """
    Extract the contents of the tarball ``tarball`` to the destination directory ``destination``.
    """
    destination.mkdir(exist_ok=True, parents=True)
    with tarfile.open(tarball) as tar:
        tar.extractall(path=destination, filter="data")



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Utility code for scripts in data.
"""

from __future__ import annotations

import json
import sys
import typing as t


def setup() -> tuple[list[str], dict[str, t.Any]]:
    """
    Fetch list of paths and potential extra configuration.

    First thing to call in an extra sanity check script in data/.
    """
    if len(sys.argv) == 3 and sys.argv[1] == "--data":
        # Preferred way: load information from JSON file
        path = sys.argv[2]
        try:
            with open(path, "rb") as f:
                data = json.load(f)
        except Exception as exc:
            raise ValueError(f"Error while reading JSON from {path}") from exc
        try:
            paths = get_list_of_strings(data, "paths")
        except ValueError as exc:
            raise ValueError(f"Invalid JSON content in {path}: {exc}") from exc
        if paths is None:
            raise ValueError(f"Broken JSON content in {path}: path is missing")
        data.pop("paths")
        return paths, data
    if len(sys.argv) >= 2:
        # It's also possible to pass a list of paths on the command line, to simplify
        # testing these scripts.
        return sys.argv[1:], {}
    # Alternatively one can pass a list of files from stdin, for example by piping
    # the output of 'git ls-files' into this script. This is also for testing these
    # scripts.
    return sys.stdin.read().splitlines(), {}


_T = t.TypeVar("_T", default=None)


@t.overload
def get_list_of_strings(
    data: dict[str, t.Any],
    key: str,
    *,
    default: None = None,
) -> list[str] | None: ...


@t.overload
def get_list_of_strings(
    data: dict[str, t.Any],
    key: str,
    *,
    default: _T,
) -> list[str] | _T: ...


def get_list_of_strings(
    data: dict[str, t.Any],
    key: str,
    *,
    default: _T | None = None,
) -> list[str] | _T | None:
    """
    Retrieves a list of strings from key ``key`` of the JSON object ``data``.

    If ``default`` is set to a list, a missing key results in this value being returned.
    """
    sentinel = object()
    value = data.get(key, sentinel)
    if value is sentinel:
        if default is not None:
            return default
        raise ValueError(f"{key!r} is not a present")
    if not isinstance(value, list):
        raise ValueError(f"{key!r} is not a list, but {type(key)}")
    if not all(isinstance(entry, str) for entry in value):
        raise ValueError(f"{key!r} is not a list of strings")
    return t.cast(list[str], value)


def get_bool(
    data: dict[str, t.Any],
    key: str,
    *,
    default: bool | None = None,
) -> bool:
    """
    Retrieves a boolean from key ``key`` of the JSON object ``data``.

    If ``default`` is set to a boolean, a missing key results in this value being returned.
    """
    sentinel = object()
    value = data.get(key, sentinel)
    if value is sentinel:
        if default is not None:
            return default
        raise ValueError(f"{key!r} is not a present")
    if not isinstance(value, bool):
        raise ValueError(f"{key!r} is not a bool, but {type(key)}")
    return value



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox extra checks session.
"""

from __future__ import annotations

from dataclasses import asdict, dataclass

import nox

from .utils import (
    compose_description,
    run_bare_script,
)


@dataclass
class ActionGroup:
    """
    Defines an action group.
    """

    # Name of the action group.
    name: str
    # Regex pattern to match modules that could belong to this action group.
    pattern: str
    # Doc fragment that members of the action group must have, but no other module
    # must have
    doc_fragment: str
    # Exclusion list of modules that match the regex, but should not be part of the
    # action group. All other modules matching the regex are assumed to be part of
    # the action group.
    exclusions: list[str] | None = None


@dataclass
class AvoidCharacterGroup:  # pylint: disable=too-many-instance-attributes
    """
    Defines an avoid character group.
    """

    # User-friendly name
    name: str | None
    # Regular expression
    regex: str
    # Extensions, paths, and directories to look for.
    # If None (not specified), will consider all files.
    match_extensions: list[str] | None
    match_paths: list[str] | None
    match_directories: list[str] | None
    # Extensions, paths, and directories to skip.
    skip_extensions: list[str]
    skip_paths: list[str]
    skip_directories: list[str]


def add_extra_checks(
    *,
    make_extra_checks_default: bool = True,
    # no-unwanted-files:
    run_no_unwanted_files: bool = True,
    no_unwanted_files_module_extensions: (
        list[str] | None
    ) = None,  # default: .cs, .ps1, .psm1, .py
    no_unwanted_files_other_extensions: list[str] | None = None,  # default: .py, .pyi
    no_unwanted_files_yaml_extensions: list[str] | None = None,  # default: .yml, .yaml
    no_unwanted_files_skip_paths: list[str] | None = None,  # default: []
    no_unwanted_files_skip_directories: list[str] | None = None,  # default: []
    no_unwanted_files_yaml_directories: (
        list[str] | None
    ) = None,  # default: plugins/test/, plugins/filter/
    no_unwanted_files_allow_symlinks: bool = False,
    # action-groups:
    run_action_groups: bool = False,
    action_groups_config: list[ActionGroup] | None = None,
    # no-trailing-whitespace:
    run_no_trailing_whitespace: bool = False,
    no_trailing_whitespace_skip_paths: list[str] | None = None,
    no_trailing_whitespace_skip_directories: list[str] | None = None,
    # avoid-character-group:
    run_avoid_characters: bool = False,
    avoid_character_group: list[AvoidCharacterGroup] | None = None,
) -> None:
    """
    Add extra-checks session for extra checks.
    """

    def execute_no_unwanted_files(session: nox.Session) -> None:
        run_bare_script(
            session,
            "no-unwanted-files",
            extra_data={
                "module_extensions": no_unwanted_files_module_extensions
                or [".cs", ".ps1", ".psm1", ".py"],
                "other_extensions": no_unwanted_files_other_extensions
                or [".py", ".pyi"],
                "yaml_extensions": no_unwanted_files_yaml_extensions
                or [".yml", ".yaml"],
                "skip_paths": no_unwanted_files_skip_paths or [],
                "skip_directories": no_unwanted_files_skip_directories or [],
                "yaml_directories": no_unwanted_files_yaml_directories
                or ["plugins/test/", "plugins/filter/"],
                "allow_symlinks": no_unwanted_files_allow_symlinks,
            },
        )

    def execute_action_groups(session: nox.Session) -> None:
        if action_groups_config is None:
            session.warn("Skipping action-groups since config is not provided...")
            return
        run_bare_script(
            session,
            "action-groups",
            extra_data={
                "config": [asdict(cfg) for cfg in action_groups_config],
            },
        )

    def execute_no_trailing_whitespace(session: nox.Session) -> None:
        run_bare_script(
            session,
            "no-trailing-whitespace",
            extra_data={
                "skip_paths": no_trailing_whitespace_skip_paths or [],
                "skip_directories": no_trailing_whitespace_skip_directories or [],
            },
        )

    def execute_avoid_characters(session: nox.Session) -> None:
        if avoid_character_group is None:
            session.warn("Skipping avoid-characters since config is not provided...")
            return
        run_bare_script(
            session,
            "avoid-characters",
            extra_data={
                "config": [asdict(cfg) for cfg in avoid_character_group],
            },
        )

    def extra_checks(session: nox.Session) -> None:
        if run_no_unwanted_files:
            execute_no_unwanted_files(session)
        if run_action_groups:
            execute_action_groups(session)
        if run_no_trailing_whitespace:
            execute_no_trailing_whitespace(session)
        if run_avoid_characters:
            execute_avoid_characters(session)

    extra_checks.__doc__ = compose_description(
        prefix={
            "one": "Run extra checker:",
            "other": "Run extra checkers:",
        },
        programs={
            "no-unwanted-files": (
                "checks for unwanted files in plugins/"
                if run_no_unwanted_files
                else False
            ),
            "action-groups": "validate action groups" if run_action_groups else False,
            "no-trailing-whitespace": (
                "avoid trailing whitespace" if run_no_trailing_whitespace else False
            ),
            "avoid-character-groups": (
                "avoid characters" if run_avoid_characters else False
            ),
        },
    )
    nox.session(
        name="extra-checks",
        python=False,
        default=make_extra_checks_default,
    )(extra_checks)


__all__ = [
    "ActionGroup",
    "add_extra_checks",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox sessions.
"""

from __future__ import annotations

import json
import os
import sys

import nox

from .collections import (  # Re-export for usage in noxfiles
    CollectionSetup,
    prepare_collections,
)
from .utils import get_registered_sessions


def add_matrix_generator() -> None:
    """
    Add a session that generates matrixes for CI systems.
    """

    def matrix_generator(
        session: nox.Session,  # pylint: disable=unused-argument
    ) -> None:
        registered_sessions = get_registered_sessions()

        json_output = os.environ.get("ANTSIBULL_NOX_MATRIX_JSON")
        if json_output:
            print(f"Writing JSON output to {json_output}...")
            with open(json_output, "wt", encoding="utf-8") as f:
                f.write(json.dumps(registered_sessions))

        github_output = os.environ.get("GITHUB_OUTPUT")
        if github_output:
            print(f"Writing GitHub output to {github_output}...")
            with open(github_output, "at", encoding="utf-8") as f:
                for name, sessions in registered_sessions.items():
                    f.write(f"{name}={json.dumps(sessions)}\n")

        for name, sessions in sorted(registered_sessions.items()):
            print(f"{name} ({len(sessions)}):")
            for session_data in sessions:
                data = session_data.copy()
                session_name = data.pop("name")
                print(f"  {session_name}: {data}")

        sys.stdout.flush()

    matrix_generator.__doc__ = "Generate matrix for CI systems."
    nox.session(
        name="matrix-generator",
        python=False,
        default=False,
    )(matrix_generator)


__all__ = [
    "CollectionSetup",
    "add_matrix_generator",
    "prepare_collections",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

import re
import urllib.error
import urllib.request

from antsibull_nox.ansible import (
    _ANSIBLE_EOL_MAX_VERSION,
    _ANSIBLE_EOL_REPO,
    _CURRENT_DEVEL_VERSION,
    _CURRENT_MILESTONE_VERSION,
)
from antsibull_nox.utils import Version

_ANSIBLE_CORE_VERSION_REGEX = re.compile(r"""__version__ = (?:'([^']+)'|"([^"]+)")""")
_EOL_ANSIBLE_BRANCH_TEST_URL = (
    "https://raw.githubusercontent.com/{repo}/refs/heads/stable-{version}/README.md"
)


def get_branch_version(branch_name: str) -> Version:
    url = (
        "https://raw.githubusercontent.com/ansible/ansible/"
        f"refs/heads/{branch_name}/lib/ansible/release.py"
    )
    release_py = urllib.request.urlopen(url).read().decode("utf-8")
    m = _ANSIBLE_CORE_VERSION_REGEX.search(release_py)
    if not m:
        raise ValueError(f"Cannot find ansible-core version in {url}:\n{release_py}")
    return Version.parse(m.group(1) or m.group(2))


def test_check_devel_version() -> None:
    assert get_branch_version("devel") == _CURRENT_DEVEL_VERSION


def test_check_milestone_version() -> None:
    assert get_branch_version("milestone") == _CURRENT_MILESTONE_VERSION


def does_exist(url: str) -> int:
    try:
        request = urllib.request.urlopen(urllib.request.Request(url, method="HEAD"))
        request.close()
        return 200 <= request.status < 300
    except urllib.error.HTTPError as exc:
        if exc.status == 404:
            return False
        raise


def test_check_eol_ansible() -> None:
    max_url = _EOL_ANSIBLE_BRANCH_TEST_URL.format(
        repo=_ANSIBLE_EOL_REPO, version=_ANSIBLE_EOL_MAX_VERSION
    )
    assert does_exist(max_url) is True

    next_version = _ANSIBLE_EOL_MAX_VERSION.next_minor_version()
    next_max_url = _EOL_ANSIBLE_BRANCH_TEST_URL.format(
        repo=_ANSIBLE_EOL_REPO, version=next_version
    )
    assert does_exist(next_max_url) is False



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Build Ansible collections.
"""

from __future__ import annotations

from pathlib import Path

import nox

from ..paths import (
    copy_collection,
    remove_path,
)
from .data import (
    CollectionData,
)
from .search import (
    load_collection_data_from_disk,
)
from .utils import (
    force_collection_version,
)


def build_collection(
    session: nox.Session,
) -> tuple[Path | None, CollectionData, str]:
    """
    Build the current collection.

    Return a tuple (path, collection_data, version), where path might be None in case
    commands are not actually run.
    """
    tmp = Path(session.create_tmp())
    collection_dir = tmp / "collection"
    remove_path(collection_dir)
    copy_collection(Path.cwd(), collection_dir)

    collection = load_collection_data_from_disk(collection_dir, accept_manifest=False)
    version = collection.version
    if not version:
        version = "0.0.1"
        force_collection_version(collection_dir, version=version)

    with session.chdir(collection_dir):
        build_ran = session.run("ansible-galaxy", "collection", "build") is not None

    tarball = (
        collection_dir / f"{collection.namespace}-{collection.name}-{version}.tar.gz"
    )
    if build_ran and not tarball.is_file():
        files = "\n".join(
            f"* {path.name}" for path in collection_dir.iterdir() if not path.is_dir()
        )
        session.error(f"Cannot find file {tarball}! List of all files:\n{files}")

    return tarball if build_ran else None, collection, version


__all__ = [
    "build_collection",
]



# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Execution environment definition generator.
"""

from __future__ import annotations

import typing as t
from copy import deepcopy
from pathlib import Path

from antsibull_fileutils.yaml import store_yaml_file

from .collection import CollectionData

ALLOWED_EE_DEFINITION_VERSIONS = {3}


def find_dict(destination: dict[str, t.Any], path: list[str]) -> dict[str, t.Any]:
    """
    Find/create dictionary determined by ``path`` in ``destination``.
    """
    for index, key in enumerate(path):
        if key not in destination:
            destination[key] = {}
        if not isinstance(destination[key], dict):
            raise ValueError(
                f"Expected a dictionary at {'.'.join(path[:index + 1])},"
                f" but found {type(destination[key])}"
            )
        destination = destination[key]
    return destination


def set_value(destination: dict[str, t.Any], path: list[str], value: t.Any) -> None:
    """
    Set value determined by ``path`` in ``destination`` to ``value``.
    """
    find_dict(destination, path[:-1])[path[-1]] = value


def generate_ee_config(
    *,
    directory: Path,
    collection_tarball_path: Path,
    collection_data: CollectionData,  # pylint: disable=unused-argument
    ee_config: dict[str, t.Any],
) -> None:
    """
    Create execution environment definition.
    """
    config = deepcopy(ee_config)

    if config.get("version") not in ALLOWED_EE_DEFINITION_VERSIONS:
        raise ValueError(f"Invalid EE definition version {config.get('version')!r}")

    # Add Galaxy requirements file
    store_yaml_file(
        directory / "requirements.yml",
        {
            "collections": [
                {
                    "name": f"src/{collection_tarball_path.name}",
                    "type": "file",
                },
            ],
        },
    )
    set_value(config, ["dependencies", "galaxy"], "requirements.yml")

    # Add collection
    if "additional_build_files" not in config:
        config["additional_build_files"] = []
    if not isinstance(config["additional_build_files"], list):
        raise ValueError(
            f"Expected a list at additional_build_files,"
            f" but found {type(config['additional_build_files'])}"
        )
    config["additional_build_files"].append(
        {
            "src": str(collection_tarball_path),
            "dest": "src",
        }
    )

    store_yaml_file(directory / "execution-environment.yml", config)


def merge(
    destination: dict[str, t.Any],
    source: dict[str, t.Any],
    *,
    path: list[str] | None = None,
    ok_paths: list[list[str]] | None = None,
    source_name: str,
) -> None:
    """
    Given two EE configs, merge the second into the first.

    In case values are specified in both configs, an error is reported.
    """
    if path is None:
        path = []
    for k, v in source.items():
        if destination.get(k) is None:
            destination[k] = v
            continue
        sub_path = path + [k]
        if not isinstance(destination[k], dict) or not isinstance(v, dict):
            if ok_paths and sub_path in ok_paths:
                destination[k] = v
                continue
            raise ValueError(
                f"Value {'.'.join(sub_path)} is already present"
                f" in the EE config (type {type(destination[k])});"
                f" cannot overwrite with value (type {type(source[k])}) from {source_name}"
            )
        merge(
            destination[k],
            source[k],
            path=sub_path,
            source_name=source_name,
            ok_paths=ok_paths,
        )


def create_ee_config(
    *,
    version: t.Literal[3],
    base_image: str | None = None,
    base_image_is_default: bool = False,
    dependencies: dict[str, t.Any] | None = None,
    config: dict[str, t.Any] | None = None,
) -> dict[str, t.Any]:
    """
    Create execution environment from parameters.
    """

    if version not in ALLOWED_EE_DEFINITION_VERSIONS:
        raise ValueError(f"Invalid EE definition version {version!r}")

    final_config: dict[str, t.Any] = {}

    # Merge session parameters
    parameter_config: dict[str, t.Any] = {
        "version": version,
    }
    if base_image is not None:
        parameter_config["images"] = {"base_image": {"name": base_image}}
    if dependencies:
        parameter_config["dependencies"] = dependencies
    merge(final_config, parameter_config, path=[], source_name="session parameters")

    # Merge passed config
    if config:
        merge(
            final_config,
            config,
            source_name="config",
            ok_paths=(
                [["images", "base_image"], ["images", "base_image", "name"]]
                if base_image_is_default
                else []
            ),
        )

    return final_config


__all__ = ["generate_ee_config", "create_ee_config"]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Path utilities.
"""

from __future__ import annotations

import atexit
import functools
import os
import shutil
from pathlib import Path

from antsibull_fileutils.copier import Copier, GitCopier
from antsibull_fileutils.tempfile import (
    ansible_mkdtemp,
    find_tempdir,
    is_acceptable_tempdir,
)
from antsibull_fileutils.vcs import detect_vcs, list_git_files


def find_data_directory() -> Path:
    """
    Retrieve the directory for antsibull_nox.data on disk.
    """
    return Path(__file__).parent / "data"


def match_path(path: str, is_file: bool, paths: list[str]) -> bool:
    """
    Check whether a path (that is a file or not) matches a given list of paths.
    """
    for check in paths:
        if check == path:
            return True
        if not is_file:
            if not check.endswith("/"):
                check += "/"
            if path.startswith(check):
                return True
    return False


def restrict_paths(paths: list[str], restrict: list[str]) -> list[str]:
    """
    Restrict a list of paths with a given set of restrictions.
    """
    result = []
    for path in paths:
        is_file = os.path.isfile(path)
        if not is_file and not path.endswith("/"):
            path += "/"
        if not match_path(path, is_file, restrict):
            if not is_file:
                for check in restrict:
                    if check.startswith(path) and os.path.exists(check):
                        result.append(check)
            continue
        result.append(path)
    return result


def _scan_remove_paths(
    path: str, remove: list[str], extensions: list[str] | None
) -> list[str]:
    result = []
    for root, dirs, files in os.walk(path, topdown=True):
        if not root.endswith("/"):
            root += "/"
        if match_path(root, False, remove):
            continue
        if all(not check.startswith(root) for check in remove):
            dirs[:] = []
            result.append(root)
            continue
        for file in files:
            if extensions and os.path.splitext(file)[1] not in extensions:
                continue
            filepath = os.path.normpath(os.path.join(root, file))
            if not match_path(filepath, True, remove):
                result.append(filepath)
        for directory in list(dirs):
            if directory == "__pycache__":
                continue
            dirpath = os.path.normpath(os.path.join(root, directory))
            if match_path(dirpath, False, remove):
                dirs.remove(directory)
                continue
    return result


def remove_paths(
    paths: list[str], remove: list[str], extensions: list[str] | None
) -> list[str]:
    """
    Restrict a list of paths by removing paths.

    If ``extensions`` is specified, only files matching this extension
    will be considered when files need to be explicitly enumerated.
    """
    result = []
    for path in paths:
        is_file = os.path.isfile(path)
        if not is_file and not path.endswith("/"):
            path += "/"
        if match_path(path, is_file, remove):
            continue
        if not is_file and any(check.startswith(path) for check in remove):
            result.extend(_scan_remove_paths(path, remove, extensions))
            continue
        result.append(path)
    return result


def filter_paths(
    paths: list[str],
    /,
    remove: list[str] | None = None,
    restrict: list[str] | None = None,
    extensions: list[str] | None = None,
) -> list[str]:
    """
    Modifies a list of paths by restricting to and/or removing paths.
    """
    if restrict:
        paths = restrict_paths(paths, restrict)
    if remove:
        paths = remove_paths(paths, remove, extensions)
    return [path for path in paths if os.path.exists(path)]


@functools.cache
def list_all_files() -> list[Path]:
    """
    List all files of interest in the repository.
    """
    directory = Path.cwd()
    vcs = detect_vcs(directory)
    if vcs == "git":
        return [directory / path.decode("utf-8") for path in list_git_files(directory)]
    result = []
    for root, dirs, files in os.walk(directory, topdown=True):
        root_path = Path(root)
        for file in files:
            result.append(root_path / file)
        if root_path == directory and ".nox" in dirs:
            dirs.remove(".nox")
    return result


def remove_path(path: Path) -> None:
    """
    Delete a path.
    """
    if not path.is_symlink() and path.is_dir():
        shutil.rmtree(path)
    elif path.exists():
        path.unlink()


def copy_collection(source: Path, destination: Path) -> None:
    """
    Copy a collection from source to destination.

    Automatically detect supported VCSs and use their information to avoid
    copying ignored files.
    """
    if destination.exists():
        remove_path(destination)
    vcs = detect_vcs(source)
    copier = {
        "none": Copier,
        "git": GitCopier,
    }.get(vcs, Copier)()
    copier.copy(source, destination, exclude_root=[".nox", ".tox"])


def _is_acceptable_tempdir(directory: Path, exclude: Path) -> bool:
    # Avoid being inside an ansible_collections tree
    if not is_acceptable_tempdir(directory):
        return False
    # Avoid something that's inside exclude
    if directory.is_relative_to(exclude):
        return False
    return True


def get_outside_temp_directory(exclude: Path | None = None) -> Path:
    """
    Find a temporary directory root that's outside the provided path.

    Raises ``ValueError`` in case no candidate was found.
    """
    if exclude is None:
        exclude = Path.cwd()
    return find_tempdir(lambda path: _is_acceptable_tempdir(path, exclude))


def create_temp_directory(basename: str) -> Path:
    """
    Create a temporary directory outside an existing ansible_collections tree.
    """
    path = ansible_mkdtemp(prefix=basename)

    def cleanup() -> None:
        remove_path(path)

    atexit.register(cleanup)
    return path


def copy_directory_tree_into(source: Path, destination: Path) -> None:
    """
    Copy the directory tree from ``source`` into the tree at ``destination``.

    If ``destination`` does not yet exist, it will be created first.
    """
    if not source.is_dir():
        return
    destination.mkdir(parents=True, exist_ok=True)
    for root, _, files in source.walk():
        path = destination / root.relative_to(source)
        path.mkdir(exist_ok=True)
        for file in files:
            dest = path / file
            remove_path(dest)
            shutil.copy2(root / file, dest, follow_symlinks=False)


__all__ = [
    "copy_collection",
    "copy_directory_tree_into",
    "create_temp_directory",
    "filter_paths",
    "find_data_directory",
    "get_outside_temp_directory",
    "list_all_files",
    "remove_path",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

# PYTHON_ARGCOMPLETE_OK

"""Entrypoint to the antsibull-docs script."""

from __future__ import annotations

import argparse
import os
import os.path
import sys
from collections.abc import Callable

from . import __version__
from .init import create_initial_config as _create_initial_config
from .lint_config import lint_config as _lint_config

try:
    import argcomplete

    HAS_ARGCOMPLETE = True
except ImportError:
    HAS_ARGCOMPLETE = False


def lint_config(_: argparse.Namespace) -> int:
    """
    Lint antsibull-nox config file.
    """
    errors = _lint_config()
    for error in errors:
        print(error)
    return 0 if len(errors) == 0 else 3


def create_initial_config(_: argparse.Namespace) -> int:
    """
    Create noxfile.py and antsibull-nox.toml.
    """
    try:
        _create_initial_config()
    except Exception as exc:  # pylint: disable=broad-exception-caught
        print(f"Error: {exc}", file=sys.stderr)
        return 3
    return 0


# Mapping from command line subcommand names to functions which implement those.
# The functions need to take a single argument, the processed list of args.
ARGS_MAP: dict[str, Callable[[argparse.Namespace], int]] = {
    "lint-config": lint_config,
    "init": create_initial_config,
}


class InvalidArgumentError(Exception):
    """
    Error while parsing arguments.
    """


def parse_args(program_name: str, args: list[str]) -> argparse.Namespace:
    """
    Parse the command line arguments.
    """

    toplevel_parser = argparse.ArgumentParser(
        prog=program_name,
        description="Script to manage generated documentation for ansible",
    )
    toplevel_parser.add_argument(
        "--version",
        action="version",
        version=__version__,
        help="Print the antsibull-nox version",
    )
    subparsers = toplevel_parser.add_subparsers(
        title="Subcommands", dest="command", help="for help use: `SUBCOMMANDS -h`"
    )
    subparsers.required = True

    subparsers.add_parser(
        "lint-config",
        description="Lint antsibull-nox configuration file",
    )

    subparsers.add_parser(
        "init",
        description="Create noxfile and antsibull-nox configuration file",
    )

    # This must come after all parser setup
    if HAS_ARGCOMPLETE:
        argcomplete.autocomplete(toplevel_parser)

    parsed_args: argparse.Namespace = toplevel_parser.parse_args(args)
    return parsed_args


def run(args: list[str]) -> int:
    """
    Run the program.
    """
    program_name = os.path.basename(args[0])
    try:
        parsed_args: argparse.Namespace = parse_args(program_name, args[1:])
    except InvalidArgumentError as e:
        print(e, file=sys.stderr)
        return 2

    return ARGS_MAP[parsed_args.command](parsed_args)


def main() -> int:
    """
    Entrypoint called from the script.

    Return codes:
        :0: Success
        :1: Unhandled error.  See the Traceback for more information.
        :2: There was a problem with the command line arguments
    """
    return run(sys.argv)


if __name__ == "__main__":
    sys.exit(main())



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""Tool to create initial configuration."""

from pathlib import Path

from .collection.search import GALAXY_YML
from .config import CONFIG_FILENAME
from .lint_config import NOXFILE_PY

NOXFILE_CONTENT = r"""
# The following metadata allows Python runners and nox to install the required
# dependencies for running this Python script:
#
# /// script
# dependencies = ["nox>=2025.02.09", "antsibull-nox"]
# ///

import sys

import nox


# We try to import antsibull-nox, and if that doesn't work, provide a more useful
# error message to the user.
try:
    import antsibull_nox
except ImportError:
    print("You need to install antsibull-nox in the same Python environment as nox.")
    sys.exit(1)


antsibull_nox.load_antsibull_nox_toml()


# Allow to run the noxfile with `python noxfile.py`, `pipx run noxfile.py`, or similar.
# Requires nox >= 2025.02.09
if __name__ == "__main__":
    nox.main()
"""

CONFIG_CONTENT = r"""
[sessions]

[sessions.lint]
# disable reformatting for now
run_isort = false

# disable reformatting for now
run_black = false

# Add more configuration settings here to adjust to your collection;
# see https://ansible.readthedocs.io/projects/antsibull-nox/config-file/#basic-linting-sessions

[sessions.docs_check]
# Add configuration settings here to adjust to your collection;
# see https://ansible.readthedocs.io/projects/antsibull-nox/config-file/#collection-documentation-check
"""


def create_initial_config(*, path: Path | None = None) -> None:
    """
    Write noxfile.py and antsibull-nox.toml for a collection.
    """
    if path is None:
        path = Path()

    galaxy_yml = path / GALAXY_YML
    if not galaxy_yml.is_file():
        raise ValueError(f"Cannot find {galaxy_yml}")

    noxfile = path / NOXFILE_PY
    config_file = path / CONFIG_FILENAME
    for file in (noxfile, config_file):
        if file.exists():
            raise ValueError(f"{file} already exists")

    noxfile.write_text(NOXFILE_CONTENT)
    config_file.write_text(CONFIG_CONTENT)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

from __future__ import annotations

import re
import typing as t

import pytest

from antsibull_nox.ee_config import (
    create_ee_config,
    set_value,
)


def re_match(value: str) -> str:
    return f"^{re.escape(value)}$"


SET_VALUE_DATA: list[tuple[dict[str, t.Any], list[str], t.Any, dict[str, t.Any]]] = [
    (
        {},
        ["foo"],
        "bar",
        {
            "foo": "bar",
        },
    ),
    (
        {},
        ["foo", "bam"],
        "boo",
        {
            "foo": {
                "bam": "boo",
            },
        },
    ),
    (
        {
            "foo": {
                "bar": "baz",
            },
        },
        ["foo", "bam"],
        "boo",
        {
            "foo": {
                "bar": "baz",
                "bam": "boo",
            },
        },
    ),
]


@pytest.mark.parametrize(
    "destination, path, value, expected",
    SET_VALUE_DATA,
)
def test_set_value(
    destination: dict[str, t.Any],
    path: list[str],
    value: t.Any,
    expected: dict[str, t.Any],
) -> None:
    set_value(destination, path, value)
    assert destination == expected


SET_VALUE_FAIL_DATA: list[tuple[dict[str, t.Any], list[str], t.Any, str]] = [
    (
        {
            "foo": [],
        },
        ["foo", "bar"],
        "bar",
        re_match("Expected a dictionary at foo, but found <class 'list'>"),
    ),
]


@pytest.mark.parametrize(
    "destination, path, value, expected_regex",
    SET_VALUE_FAIL_DATA,
)
def test_set_value_fail(
    destination: dict[str, t.Any],
    path: list[str],
    value: t.Any,
    expected_regex: str,
) -> None:
    with pytest.raises(ValueError, match=expected_regex):
        set_value(destination, path, value)


CREATE_EE_CONFIG_DATA: list[
    tuple[
        int,
        str | None,
        bool,
        dict[str, t.Any] | None,
        dict[str, t.Any] | None,
        dict[str, t.Any],
    ]
] = [
    (
        3,
        "foo-bar",
        False,
        None,
        None,
        {
            "images": {
                "base_image": {
                    "name": "foo-bar",
                },
            },
            "version": 3,
        },
    ),
    (
        3,
        "foo-bar",
        True,
        {
            "foo": "bar",
        },
        {
            "images": {
                "base_image": {
                    "name": "baz-bam",
                },
            },
        },
        {
            "dependencies": {
                "foo": "bar",
            },
            "images": {
                "base_image": {
                    "name": "baz-bam",
                },
            },
            "version": 3,
        },
    ),
]


@pytest.mark.parametrize(
    "version, base_image, base_image_is_default, dependencies, config, expected",
    CREATE_EE_CONFIG_DATA,
)
def test_create_ee_config(
    version: int,
    base_image: str | None,
    base_image_is_default: bool,
    dependencies: dict[str, t.Any] | None,
    config: dict[str, t.Any] | None,
    expected: dict[str, t.Any],
) -> None:
    actual = create_ee_config(
        version=version,  # type: ignore
        base_image=base_image,
        base_image_is_default=base_image_is_default,
        dependencies=dependencies,
        config=config,
    )
    assert actual == expected


CREATE_EE_CONFIG_FAIL_DATA: list[
    tuple[int, str | None, bool, dict[str, t.Any] | None, dict[str, t.Any] | None, str]
] = [
    (
        3,
        None,
        False,
        None,
        {
            "version": 3,
        },
        re_match(
            "Value version is already present in the EE config (type <class 'int'>);"
            " cannot overwrite with value (type <class 'int'>) from config"
        ),
    ),
    (
        3,
        "foo-bar",
        False,
        None,
        {
            "images": {
                "base_image": {
                    "name": "baz-bam",
                },
            },
        },
        re_match(
            "Value images.base_image.name is already present in the EE config (type <class 'str'>);"
            " cannot overwrite with value (type <class 'str'>) from config"
        ),
    ),
    (
        3,
        "foo-bar",
        False,
        None,
        {
            "images": [],
        },
        re_match(
            "Value images is already present in the EE config (type <class 'dict'>);"
            " cannot overwrite with value (type <class 'list'>) from config"
        ),
    ),
    (
        -1,
        None,
        False,
        None,
        {},
        re_match("Invalid EE definition version -1"),
    ),
]


@pytest.mark.parametrize(
    "version, base_image, base_image_is_default, dependencies, config, expected_regex",
    CREATE_EE_CONFIG_FAIL_DATA,
)
def test_create_ee_config_fail(
    version: int,
    base_image: str | None,
    base_image_is_default: bool,
    dependencies: dict[str, t.Any] | None,
    config: dict[str, t.Any] | None,
    expected_regex: str,
) -> None:
    with pytest.raises(ValueError, match=expected_regex):
        create_ee_config(
            version=version,  # type: ignore
            base_image=base_image,
            base_image_is_default=base_image_is_default,
            dependencies=dependencies,
            config=config,
        )



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Data types for collections.
"""

from __future__ import annotations

import base64
import hashlib
from dataclasses import dataclass
from pathlib import Path


@dataclass
class CollectionData:  # pylint: disable=too-many-instance-attributes
    """
    An Ansible collection.
    """

    collections_root_path: Path | None
    path: Path
    namespace: str
    name: str
    full_name: str
    version: str | None
    dependencies: dict[str, str]
    current: bool

    @classmethod
    def create(
        cls,
        *,
        collections_root_path: Path | None = None,
        path: Path,
        full_name: str,
        version: str | None = None,
        dependencies: dict[str, str] | None = None,
        current: bool = False,
    ):
        """
        Create a CollectionData object.
        """
        namespace, name = full_name.split(".", 1)
        return CollectionData(
            collections_root_path=collections_root_path,
            path=path,
            namespace=namespace,
            name=name,
            full_name=full_name,
            version=version,
            dependencies=dependencies or {},
            current=current,
        )


@dataclass
class SetupResult:
    """
    Information on how the collections are set up.
    """

    # The path of the ansible_collections directory.
    root: Path

    # Data on the current collection (as in the repository).
    current_collection: CollectionData

    # If it was installed, the path of the current collection inside the collection tree below root.
    current_path: Path | None


@dataclass
class CollectionSource:
    """
    Collection installation source.
    """

    # The collection's name
    name: str

    # The collection's installation source (can be passed to 'ansible-galaxy collection install')
    source: str

    @staticmethod
    def parse(collection_name: str, source: str | CollectionSource) -> CollectionSource:
        """
        Parse a CollectionSource object.
        """
        if isinstance(source, str):
            return CollectionSource(name=collection_name, source=source)

        if source.name != collection_name:
            raise ValueError(
                f"Collection name should be {collection_name!r}, but is {source.name!r}"
            )
        return source

    def identifier(self) -> str:
        """
        Compute a source identifier.
        """
        hasher = hashlib.sha256()
        hasher.update(self.name.encode("utf-8"))
        hasher.update(b"::")
        hasher.update(self.source.encode("utf-8"))
        return base64.b32encode(hasher.digest())[:16].decode("ascii")


__all__ = [
    "CollectionData",
    "SetupResult",
    "CollectionSource",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

# pylint: disable=missing-function-docstring

"""
Helpers for testing.
"""

from __future__ import annotations

import contextlib
import os
from pathlib import Path


@contextlib.contextmanager
def chdir(dir: Path):
    current = Path.cwd()
    try:
        os.chdir(dir)
        yield
    finally:
        os.chdir(current)



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

from __future__ import annotations

import pytest

from antsibull_nox.utils import Version, version_range

VERSION_PARSE_DATA: list[tuple[str, Version]] = [
    (
        "1.2",
        Version(1, 2),
    ),
    (
        "1.2.3b1",
        Version(1, 2),
    ),
]


@pytest.mark.parametrize(
    "version_string, expected_version",
    VERSION_PARSE_DATA,
)
def test_version_parse(version_string: str, expected_version: Version) -> None:
    assert Version.parse(version_string) == expected_version


def test_verison_parse_fail() -> None:
    with pytest.raises(ValueError, match=r"^Cannot parse '1' as version string\.$"):
        Version.parse("1")

    with pytest.raises(ValueError, match=r"^Cannot parse '1.a' as version string.$"):
        Version.parse("1.a")


def test_version_next() -> None:
    assert Version(2, 3).next_minor_version() == Version(2, 4)
    assert Version(1, 9).next_minor_version() == Version(1, 10)


def test_version_prev() -> None:
    assert Version(2, 4).previous_minor_version() == Version(2, 3)
    assert Version(1, 10).previous_minor_version() == Version(1, 9)

    with pytest.raises(ValueError, match="^No previous minor version exists$"):
        Version(1, 0).previous_minor_version()


VERSION_RANGE_DATA: list[tuple[str, str, bool, list[str]]] = [
    (
        "1.2",
        "1.3",
        True,
        ["1.2", "1.3"],
    ),
    (
        "1.2",
        "1.3",
        False,
        ["1.2"],
    ),
]


@pytest.mark.parametrize(
    "start, end, inclusive, expected_versions",
    VERSION_RANGE_DATA,
)
def test_version_range(
    start: str, end: str, inclusive: bool, expected_versions: list[str]
) -> None:
    sv = Version.parse(start)
    ev = Version.parse(end)
    evs = [Version.parse(v) for v in expected_versions]
    versions = [v for v in version_range(sv, ev, inclusive=inclusive)]
    assert versions == evs


def test_version_range_fail() -> None:
    with pytest.raises(
        ValueError,
        match=(
            r"^Cannot list all versions from 1\.2 to 2\.3"
            r" since they have different major versions$"
        ),
    ):
        next(version_range(Version(1, 2), Version(2, 3), inclusive=True))



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""
Create nox ansible-lint session.
"""

from __future__ import annotations

import nox

from .collections import prepare_collections
from .utils import install


def add_ansible_lint(
    *,
    make_ansible_lint_default: bool = True,
    ansible_lint_package: str = "ansible-lint",
    strict: bool = False,
) -> None:
    """
    Add a session that runs ansible-lint.
    """

    def compose_dependencies() -> list[str]:
        return [ansible_lint_package]

    def ansible_lint(session: nox.Session) -> None:
        install(session, *compose_dependencies())
        prepared_collections = prepare_collections(
            session,
            install_in_site_packages=False,
            install_out_of_tree=True,
            extra_deps_files=["tests/integration/requirements.yml"],
        )
        if not prepared_collections:
            session.warn("Skipping ansible-lint...")
            return
        env = {"ANSIBLE_COLLECTIONS_PATH": f"{prepared_collections.current_place}"}
        command = ["ansible-lint", "--offline"]
        if strict:
            command.append("--strict")
        if session.posargs:
            command.extend(session.posargs)
        session.run(*command, env=env)

    ansible_lint.__doc__ = "Run ansible-lint."
    nox.session(
        name="ansible-lint",
        default=make_ansible_lint_default,
    )(ansible_lint)


__all__ = [
    "add_ansible_lint",
]



# Author: Felix Fontein <felix@fontein.de>
# GNU General Public License v3.0+ (see LICENSES/GPL-3.0-or-later.txt or
# https://www.gnu.org/licenses/gpl-3.0.txt)
# SPDX-License-Identifier: GPL-3.0-or-later
# SPDX-FileCopyrightText: 2025, Ansible Project

"""Lint antsibull-nox config."""

from __future__ import annotations

import ast
from pathlib import Path

from .config import lint_config_toml

NOXFILE_PY = "noxfile.py"


def _is_antsibull_nox_module(module_name: str) -> bool:
    return module_name == "antsibull_nox" or module_name.startswith("antsibull_nox.")


class _Walker:
    def __init__(self, path: Path) -> None:
        self.path = path
        self.imports: dict[str, str] = {}
        self.has_config_load = False

    @staticmethod
    def _get_name(what: ast.expr) -> str | None:
        if isinstance(what, ast.Name):
            return what.id
        if isinstance(what, ast.Attribute):
            v = _Walker._get_name(what.value)
            return None if v is None else f"{v}.{what.attr}"
        return None

    def _check_expression(self, expression: ast.expr) -> None:
        if isinstance(expression, ast.Call):
            func = _Walker._get_name(expression.func)
            if func is not None:
                func = self.imports.get(func, func)
                if "." in func:
                    module, module_func = func.rsplit(".", 1)
                    if module in self.imports:
                        func = f"{self.imports[module]}.{module_func}"
                if func == "antsibull_nox.load_antsibull_nox_toml":
                    self.has_config_load = True

    def _walk(self, statements: list[ast.stmt]) -> None:
        for statement in statements:
            # Handle imports
            if isinstance(statement, ast.Import):
                for alias in statement.names:
                    if _is_antsibull_nox_module(alias.name):
                        self.imports[alias.asname or alias.name] = alias.name
            if isinstance(statement, ast.ImportFrom):
                if statement.level == 0 and _is_antsibull_nox_module(
                    statement.module or ""
                ):
                    for alias in statement.names:
                        self.imports[alias.asname or alias.name] = (
                            f"{statement.module}.{alias.name}"
                        )
            # Handle try block
            if isinstance(statement, ast.Try):
                self._walk(statement.body)
            # Handle expressions
            if isinstance(statement, ast.Expr):
                self._check_expression(statement.value)

    def walk(self, module: ast.Module) -> list[str]:
        """
        Walk the noxfile's module and return a list of errors.
        """
        self._walk(module.body)
        errors = []
        if not self.imports:
            errors.append(f"{self.path}: Found no antsibull_nox import")
        if not self.has_config_load:
            errors.append(
                f"{self.path}: Found no call to antsibull_nox.load_antsibull_nox_toml()"
            )
        return errors


def lint_noxfile() -> list[str]:
    """
    Do basic validation for noxfile.py. Return a list of errors.
    """
    path = Path(NOXFILE_PY)
    try:
        with open(path, "rt", encoding="utf-8") as f:
            root = ast.parse(f.read(), filename=path)
    except FileNotFoundError:
        return [f"{path}: File does not exist"]
    except Exception as exc:  # pylint: disable=broad-exception-caught
        return [f"{path}:Error while parsing Python code: {exc}"]

    walker = _Walker(path)
    return walker.walk(root)


def lint_config() -> list[str]:
    """
    Lint antsibull-nox config file.
    """
    errors = lint_config_toml()
    errors.extend(lint_noxfile())
    return sorted(errors)


__all__ = ["lint_config"]
